"""
Chimera data structures



"""

import os
import re

import numpy as np

from yt.data_objects.index_subobjects.unstructured_mesh import SemiStructuredMesh
from yt.data_objects.static_output import Dataset
from yt.geometry.api import Geometry
from yt.geometry.geometry_handler import YTDataChunk
from yt.geometry.unstructured_mesh_handler import UnstructuredIndex
from yt.utilities.file_handler import HDF5FileHandler
from yt.utilities.io_handler import io_registry
from yt.utilities.logger import ytLogger as mylog
from yt.utilities.on_demand_imports import _h5py as h5py

from .fields import ChimeraFieldInfo


class ChimeraMesh(SemiStructuredMesh):
    _index_offset = 0

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


def _find_files(filename_c):
    # Returns a list of all files that share a frame number with the input
    dirname, file = os.path.split(filename_c)
    match = re.match(r"chimera_\d+_grid", file)
    if match is None:
        raise RuntimeError(
            rf"Expected filename to be of form 'chimera_\d+_grid_*', got {file!r}"
        )
    prefix = match.group()
    frames = [f for f in os.listdir(dirname) if f.startswith(prefix)]
    index_filenames = [os.path.join(dirname, f) for f in sorted(frames)]
    return index_filenames


class ChimeraUNSIndex(UnstructuredIndex):
    def __init__(self, ds, dataset_type="chimera"):
        self._handle = ds._handle
        super().__init__(ds, dataset_type)
        self.directory = os.path.dirname(self.dataset.filename)
        self.dataset_type = dataset_type

    def _initialize_mesh(self):
        self.meshes = []
        index_filenames = _find_files(
            self.dataset.filename
        )  # Retrieves list of all datafiles with the same frame number
        # Detects Yin-Yang data format
        yy = any("grid_2" in file for file in index_filenames)
        for n, file in enumerate(index_filenames):
            with h5py.File(file, "r") as f:
                nmx, nmy, nmz = tuple(f["mesh"]["array_dimensions"][:])
                l = (
                    int(file[-5:-3]) - 1
                )  # Pulls the subgrid number from the data file name
                if nmz > 2:
                    k = f["fluid"]["entropy"].shape[0]
                    r = f["mesh"]["x_ef"][:-2]
                    theta = f["mesh"]["y_ef"][:]
                    phi = f["mesh"]["z_ef"][
                        k * l : k * (l + 1) + 1
                    ]  # Pulls only the individual subgrid's band of phi values
                elif f["mesh"]["z_ef"][-1] == f["mesh"]["z_ef"][0]:
                    r = f["mesh"]["x_ef"][
                        : f["mesh"]["radial_index_bound"][1]
                        - f["mesh"]["x_ef"].shape[0]
                    ]
                    theta = f["mesh"]["y_ef"][:]
                    phi = np.array([f["mesh"]["z_ef"][0], 2 * np.pi])
                else:
                    r = f["mesh"]["x_ef"][
                        : f["mesh"]["radial_index_bound"][1]
                        - f["mesh"]["x_ef"].shape[0]
                    ]
                    theta = f["mesh"]["y_ef"][:]
                    phi = f["mesh"]["z_ef"][:]

                # Creates variables to hold the size of dimensions
                nxd = r.size
                nyd = theta.size
                nzd = phi.size
                nyzd = nyd * nzd
                nyzd_ = (nyd - 1) * (nzd - 1)

                # Generates and fills coordinate array
                coords = np.zeros((nxd, nyd, nzd, 3), dtype="float64", order="C")
                coords[:, :, :, 0] = r[:, None, None]
                coords[:, :, :, 1] = theta[None, :, None]
                coords[:, :, :, 2] = phi[None, None, :]

                if yy:
                    mylog.warning(
                        "Yin-Yang File Detected; This data is not currently supported."
                    )
                coords.shape = (nxd * nyd * nzd, 3)
                # Connectivity is an array of rows, each of which corresponds to a grid cell.
                # The 8 elements of each row are integers representing the cell vertices.
                # These integers reference the numerical index of the element of the
                # "coords" array which corresponds to the spatial coordinate.

                connectivity = np.zeros(
                    ((nyd - 1) * (nxd - 1) * (nzd - 1), 8), dtype="int64", order="C"
                )  # Creates scaffold array
                connectivity[0] = [
                    0,
                    1,
                    nzd,
                    (nzd + 1),
                    (nyzd),
                    (nyzd + 1),
                    (nyzd + nzd),
                    (nyzd + nzd + 1),
                ]  # Manually defines first coordinate set

                for p in range(
                    nzd - 1
                ):  # Increments first row around phi to define an arc of cells
                    if p > 0:
                        connectivity[p] = connectivity[p - 1] + 1
                for t in range(
                    nyd - 1
                ):  # Increments this arc around theta to define a shell
                    if t > 0:
                        connectivity[t * (nzd - 1) : (t + 1) * (nzd - 1)] = (
                            connectivity[(t - 1) * (nzd - 1) : t * (nzd - 1)] + nzd
                        )
                for r in range(
                    nxd - 1
                ):  # Increments this shell along r to define a sphere
                    if r > 0:
                        connectivity[r * (nyzd_) : (r + 1) * (nyzd_)] = (
                            connectivity[(r - 1) * (nyzd_) : r * (nyzd_)] + nyzd
                        )

                mesh = ChimeraMesh(
                    n, self.index_filename, connectivity, coords, self
                )  # Creates a mesh object

                if "grid_" in file:
                    mylog.info("Mesh %s generated", (n + 1) / len(index_filenames))
                    self.meshes.append(
                        mesh
                    )  # Adds new mesh to the list of generated meshes

    def _detect_output_fields(self):  # Reads in the available data fields
        with h5py.File(self.index_filename, "r") as f:
            fluids = [
                ("chimera", i)
                for i in f["fluid"]
                if np.shape(f["fluid"][i]) == np.shape(f["fluid"]["rho_c"])
            ]
            abundance = [
                ("chimera", i)
                for i in f["abundance"]
                if np.shape(f["abundance"][i]) == np.shape(f["fluid"]["rho_c"])
            ]
            e_rms = [("chimera", f"e_rms_{i+1}") for i in range(4)]
            lumin = [("chimera", f"lumin_{i+1}") for i in range(4)]
            num_lumin = [("chimera", f"num_lumin_{i+1}") for i in range(4)]
            a_name = [
                ("chimera", i.decode("utf-8").strip()) for i in f["abundance"]["a_name"]
            ]
            self.field_list = (
                fluids
                + abundance
                + e_rms
                + lumin
                + num_lumin
                + [("chimera", "abar")]
                + a_name
            )
            if np.shape(f["abundance"]["nse_c"]) != np.shape(f["fluid"]["rho_c"]):
                self.field_list += [("chimera", "nse_c")]

    def _chunk_io(self, dobj, cache=True, local_only=False):  # Creates Data chunk
        gobjs = getattr(dobj._current_chunk, "objs", dobj._chunk_info)
        for subset in gobjs:
            yield YTDataChunk(
                dobj, "io", [subset], self._count_selection(dobj, [subset]), cache=cache
            )

    def _setup_data_io(self):
        self.io = io_registry[self.dataset_type](self.dataset)


class ChimeraDataset(Dataset):
    _index_class = ChimeraUNSIndex  # ChimeraHierarchy
    _field_info_class = ChimeraFieldInfo

    def __init__(
        self,
        filename,
        dataset_type="chimera",
        storage_filename=None,
        units_override=None,
    ):
        # refinement factor between a grid and its subgrid
        self.refine_by = 1  # Somewhat superfluous for Chimera, but left to avoid errors

        self.fluid_types += ("chimera",)
        super().__init__(filename, dataset_type, units_override=units_override)
        self.storage_filename = storage_filename
        self._handle = HDF5FileHandler(filename)

    def _set_code_unit_attributes(self):
        # This is where quantities are created that represent the various
        # on-disk units.  These are the currently available quantities which
        # should be set, along with examples of how to set them to standard
        # values.
        #
        self.length_unit = self.quan(1.0, "cm")
        self.mass_unit = self.quan(1.0, "g")
        self.time_unit = self.quan(1.0, "s")
        self.time_unit = self.quan(1.0, "s")
        self.velocity_unit = self.quan(1.0, "cm/s")
        self.magnetic_unit = self.quan(1.0, "gauss")

    def _parse_parameter_file(self):
        with h5py.File(self.parameter_filename, "r") as f:
            # Reads in simulation time, number of dimensions and shape
            self.current_time = f["mesh"]["time"][()]
            self.dimensionality = 3
            self.domain_dimensions = f["mesh"]["array_dimensions"][()]

            self.geometry = Geometry.SPHERICAL  # Uses default spherical geometry
            self._periodicity = (False, False, True)
            dle = [
                f["mesh"]["x_ef"][0],
                f["mesh"]["y_ef"][0],
                f["mesh"]["z_ef"][0],
            ]

            if (
                self.domain_dimensions[2] <= 2
                and f["mesh"]["z_ef"][-1] == f["mesh"]["z_ef"][0]
            ):
                dre = [
                    f["mesh"]["x_ef"][
                        f["mesh"]["radial_index_bound"][1] - f["mesh"]["x_ef"].shape[0]
                    ],
                    f["mesh"]["y_ef"][-1],
                    2 * np.pi,
                ]
            else:
                dre = [
                    f["mesh"]["x_ef"][
                        f["mesh"]["radial_index_bound"][1] - f["mesh"]["x_ef"].shape[0]
                    ],
                    f["mesh"]["y_ef"][-1],
                    f["mesh"]["z_ef"][-1],
                ]
            # Sets left and right bounds based on earlier definitions
            self.domain_right_edge = np.array(dre)
            self.domain_left_edge = np.array(dle)

        self.cosmological_simulation = 0  # Chimera is not a cosmological simulation

    @classmethod
    def _is_valid(cls, *args, **kwargs):
        # This accepts a filename or a set of arguments and returns True or
        # False depending on if the file is of the type requested.
        try:
            fileh = HDF5FileHandler(args[0])
            if (
                "fluid" in fileh
                and "agr_c" in fileh["fluid"].keys()
                and "grav_x_c" in fileh["fluid"].keys()
            ):
                return True  # Numpy bless
        except (OSError, ImportError):
            pass
        return False



from numpy.testing import assert_raises

from yt.testing import assert_allclose_units, requires_file
from yt.units import YTQuantity
from yt.utilities.answer_testing.framework import data_dir_load

khi_cartesian_2D = "amrvac/kh_2d0000.dat"

# Tests for units: check that overriding certain units yields the correct derived units.
# The following are the correct normalisations
# based on length, numberdensity and temperature
length_unit = (1e9, "cm")
numberdensity_unit = (1e9, "cm**-3")
temperature_unit = (1e6, "K")
density_unit = (2.341670657200000e-15, "g*cm**-3")
mass_unit = (2.341670657200000e12, "g")
velocity_unit = (1.164508387441102e07, "cm*s**-1")
pressure_unit = (3.175492240000000e-01, "dyn*cm**-2")
time_unit = (8.587314705370271e01, "s")
magnetic_unit = (1.997608879907716, "gauss")


def _assert_normalisations_equal(ds):
    assert_allclose_units(ds.length_unit, YTQuantity(*length_unit))
    assert_allclose_units(ds.temperature_unit, YTQuantity(*temperature_unit))
    assert_allclose_units(ds.density_unit, YTQuantity(*density_unit))
    assert_allclose_units(ds.mass_unit, YTQuantity(*mass_unit))
    assert_allclose_units(ds.velocity_unit, YTQuantity(*velocity_unit))
    assert_allclose_units(ds.pressure_unit, YTQuantity(*pressure_unit))
    assert_allclose_units(ds.time_unit, YTQuantity(*time_unit))
    assert_allclose_units(ds.magnetic_unit, YTQuantity(*magnetic_unit))


@requires_file(khi_cartesian_2D)
def test_normalisations_length_temp_nb():
    # overriding length, temperature, numberdensity
    overrides = {
        "length_unit": length_unit,
        "temperature_unit": temperature_unit,
        "numberdensity_unit": numberdensity_unit,
    }
    ds = data_dir_load(khi_cartesian_2D, kwargs={"units_override": overrides})
    _assert_normalisations_equal(ds)


@requires_file(khi_cartesian_2D)
def test_normalisations_length_temp_mass():
    # overriding length, temperature, mass
    overrides = {
        "length_unit": length_unit,
        "temperature_unit": temperature_unit,
        "mass_unit": mass_unit,
    }
    ds = data_dir_load(khi_cartesian_2D, kwargs={"units_override": overrides})
    _assert_normalisations_equal(ds)


@requires_file(khi_cartesian_2D)
def test_normalisations_length_time_mass():
    # overriding length, time, mass
    overrides = {
        "length_unit": length_unit,
        "time_unit": time_unit,
        "mass_unit": mass_unit,
    }
    ds = data_dir_load(khi_cartesian_2D, kwargs={"units_override": overrides})
    _assert_normalisations_equal(ds)


@requires_file(khi_cartesian_2D)
def test_normalisations_length_vel_nb():
    # overriding length, velocity, numberdensity
    overrides = {
        "length_unit": length_unit,
        "velocity_unit": velocity_unit,
        "numberdensity_unit": numberdensity_unit,
    }
    ds = data_dir_load(khi_cartesian_2D, kwargs={"units_override": overrides})
    _assert_normalisations_equal(ds)


@requires_file(khi_cartesian_2D)
def test_normalisations_length_vel_mass():
    # overriding length, velocity, mass
    overrides = {
        "length_unit": length_unit,
        "velocity_unit": velocity_unit,
        "mass_unit": mass_unit,
    }
    ds = data_dir_load(khi_cartesian_2D, kwargs={"units_override": overrides})
    _assert_normalisations_equal(ds)


@requires_file(khi_cartesian_2D)
def test_normalisations_default():
    # test default normalisations, without overrides
    ds = data_dir_load(khi_cartesian_2D)
    assert_allclose_units(ds.length_unit, YTQuantity(1, "cm"))
    assert_allclose_units(ds.temperature_unit, YTQuantity(1, "K"))
    assert_allclose_units(
        ds.density_unit, YTQuantity(2.341670657200000e-24, "g*cm**-3")
    )
    assert_allclose_units(ds.mass_unit, YTQuantity(2.341670657200000e-24, "g"))
    assert_allclose_units(
        ds.velocity_unit, YTQuantity(1.164508387441102e04, "cm*s**-1")
    )
    assert_allclose_units(
        ds.pressure_unit, YTQuantity(3.175492240000000e-16, "dyn*cm**-2")
    )
    assert_allclose_units(ds.time_unit, YTQuantity(8.587314705370271e-05, "s"))
    assert_allclose_units(ds.magnetic_unit, YTQuantity(6.316993934686148e-08, "gauss"))


@requires_file(khi_cartesian_2D)
def test_normalisations_too_many_args():
    # test forbidden case: too many arguments (max 3 are allowed)
    overrides = {
        "length_unit": length_unit,
        "numberdensity_unit": numberdensity_unit,
        "temperature_unit": temperature_unit,
        "time_unit": time_unit,
    }
    with assert_raises(ValueError):
        data_dir_load(khi_cartesian_2D, kwargs={"units_override": overrides})


@requires_file(khi_cartesian_2D)
def test_normalisations_vel_and_length():
    # test forbidden case: both velocity and temperature are specified as overrides
    overrides = {
        "length_unit": length_unit,
        "velocity_unit": velocity_unit,
        "temperature_unit": temperature_unit,
    }
    with assert_raises(ValueError):
        data_dir_load(khi_cartesian_2D, kwargs={"units_override": overrides})



import numpy as np

import yt

ds = yt.load("IsolatedGalaxy/galaxy0030/galaxy0030")

# We start by building a default volume rendering scene

im, sc = yt.volume_render(ds, field=("gas", "density"), fname="v0.png", sigma_clip=6.0)

sc.camera.set_width(ds.arr(0.1, "code_length"))
tf = sc.get_source().transfer_function
tf.clear()
tf.add_layers(
    4, 0.01, col_bounds=[-27.5, -25.5], alpha=np.logspace(-3, 0, 4), colormap="RdBu_r"
)
sc.save("v1.png", sigma_clip=6.0)

# In this case, the default alphas used (np.logspace(-3,0,Nbins)) does not
# accentuate the outer regions of the galaxy. Let's start by bringing up the
# alpha values for each contour to go between 0.1 and 1.0

tf = sc.get_source().transfer_function
tf.clear()
tf.add_layers(
    4, 0.01, col_bounds=[-27.5, -25.5], alpha=np.logspace(0, 0, 4), colormap="RdBu_r"
)
sc.save("v2.png", sigma_clip=6.0)

# Now let's set the grey_opacity to True.  This should make the inner portions
# start to be obscured

tf.grey_opacity = True
sc.save("v3.png", sigma_clip=6.0)

# That looks pretty good, but let's start bumping up the opacity.

tf.clear()
tf.add_layers(
    4,
    0.01,
    col_bounds=[-27.5, -25.5],
    alpha=10.0 * np.ones(4, dtype="float64"),
    colormap="RdBu_r",
)
sc.save("v4.png", sigma_clip=6.0)

# Let's bump up again to see if we can obscure the inner contour.

tf.clear()
tf.add_layers(
    4,
    0.01,
    col_bounds=[-27.5, -25.5],
    alpha=30.0 * np.ones(4, dtype="float64"),
    colormap="RdBu_r",
)
sc.save("v5.png", sigma_clip=6.0)

# Now we are losing sight of everything.  Let's see if we can obscure the next
# layer

tf.clear()
tf.add_layers(
    4,
    0.01,
    col_bounds=[-27.5, -25.5],
    alpha=100.0 * np.ones(4, dtype="float64"),
    colormap="RdBu_r",
)
sc.save("v6.png", sigma_clip=6.0)

# That is very opaque!  Now lets go back and see what it would look like with
# grey_opacity = False

tf.grey_opacity = False
sc.save("v7.png", sigma_clip=6.0)

# That looks pretty different, but the main thing is that you can see that the
# inner contours are somewhat visible again.



# distutils: libraries = STD_LIBS
# distutils: sources = yt/utilities/lib/cykdtree/c_utils.cpp
# distutils: depends = yt/utilities/lib/cykdtree/c_utils.hpp
# distutils: language = c++
# distutils: extra_compile_args = CPP11_FLAG
import numpy as np

cimport numpy as np
from libc.stdint cimport int64_t, uint32_t, uint64_t
from libcpp cimport bool as cbool


def py_max_pts(np.ndarray[np.float64_t, ndim=2] pos):
    r"""Get the maximum of points along each coordinate.

    Args:
        pos (np.ndarray of float64): (n,m) array of n m-D coordinates.

    Returns:
        np.ndarray of float64: Maximum of pos along each coordinate.

    """
    cdef uint64_t n = <uint64_t>pos.shape[0]
    cdef uint32_t m = <uint32_t>pos.shape[1]
    cdef np.float64_t* cout = max_pts(&pos[0,0], n, m)
    cdef uint32_t i = 0
    cdef np.ndarray[np.float64_t] out = np.zeros(m, 'float64')
    for i in range(m):
        out[i] = cout[i]
    #free(cout)
    return out

def py_min_pts(np.ndarray[np.float64_t, ndim=2] pos):
    r"""Get the minimum of points along each coordinate.

    Args:
        pos (np.ndarray of float64): (n,m) array of n m-D coordinates.

    Returns:
        np.ndarray of float64: Minimum of pos along each coordinate.

    """
    cdef uint64_t n = <uint64_t>pos.shape[0]
    cdef uint32_t m = <uint32_t>pos.shape[1]
    cdef np.float64_t* cout = min_pts(&pos[0,0], n, m)
    cdef uint32_t i = 0
    cdef np.ndarray[np.float64_t] out = np.zeros(m, 'float64')
    for i in range(m):
        out[i] = cout[i]
    #free(cout)
    return out

def py_argmax_pts_dim(np.ndarray[np.float64_t, ndim=2] pos,
                      uint64_t[:] idx,
                      np.uint32_t d, int Lidx0, int Ridx0):
    r"""Get the maximum of points along one dimension for a subset of the
    point indices. This is essentially max(pos[idx[Lidx:(Ridx+1)], d]).

    Args:
        pos (np.ndarray of float64): (n,m) array of n m-D coordinates.
        idx (np.ndarray of uint64_t): (n,) array of indices for positions.
        d (uint32_t): Dimension to compute maximum along.
        Lidx (int): Index in idx that search should begin at.
        Ridx (int): Index in idx that search should end at.

    Returns:
        uint64_t: Index in idx that provides maximum position in the subset
            indices along dimension d.

    """
    cdef np.intp_t n = pos.shape[0]
    cdef uint32_t m = <uint32_t>pos.shape[1]
    cdef uint64_t Lidx = 0
    cdef uint64_t Ridx = <uint64_t>(n-1)
    if (Lidx0 < 0):
        Lidx = <uint64_t>(n + Lidx0)
    elif Lidx0 >= n:
        raise Exception("Left index (%d) exceeds size of positions array (%d)."
                            % (Lidx0, n))
    else:
        Lidx = <uint64_t>Lidx0
    if (Ridx0 < 0):
        Ridx = <uint64_t>(n + Ridx0)
    elif Ridx0 >= n:
        raise Exception("Right index (%d) exceeds size of positions array (%d)."
                            % (Ridx0, n))
    else:
        Ridx = <uint64_t>Ridx0
    cdef np.uint64_t cout = Lidx
    if (Ridx > Lidx):
        cout = argmax_pts_dim(&pos[0,0], &idx[0], m, d, Lidx, Ridx)
    return cout

def py_argmin_pts_dim(np.ndarray[np.float64_t, ndim=2] pos,
                      uint64_t[:] idx,
                      np.uint32_t d, int Lidx0, int Ridx0):
    r"""Get the minimum of points along one dimension for a subset of the
    point indices. This is essentially min(pos[idx[Lidx:(Ridx+1)], d]).

    Args:
        pos (np.ndarray of float64): (n,m) array of n m-D coordinates.
        idx (np.ndarray of uint64_t): (n,) array of indices for positions.
        d (uint32_t): Dimension to compute minimum along.
        Lidx (int): Index in idx that search should begin at.
        Ridx (int): Index in idx that search should end at.

    Returns:
        uint64_t: Index in idx that provides minimum position in the subset
            indices along dimension d.

    """
    cdef uint64_t n = <uint64_t>pos.shape[0]
    cdef uint32_t m = <uint32_t>pos.shape[1]
    cdef uint64_t Lidx = 0
    cdef uint64_t Ridx = n
    if (Lidx0 < 0):
        Lidx = <uint64_t>(<int>n + Lidx0)
    else:
        Lidx = <uint64_t>Lidx0
    if (Ridx0 < 0):
        Ridx = <uint64_t>(<int>n + Ridx0)
    else:
        Ridx = <uint64_t>Ridx0
    cdef np.uint64_t cout = Lidx
    if (Ridx > Lidx):
        cout = argmin_pts_dim(&pos[0,0], &idx[0], m, d, Lidx, Ridx)
    return cout

def py_quickSort(np.ndarray[np.float64_t, ndim=2] pos, np.uint32_t d):
    r"""Get the indices required to sort coordinates along one dimension.

    Args:
        pos (np.ndarray of float64): (n,m) array of n m-D coordinates.
        d (np.uint32_t): Dimension that pos should be sorted along.

    Returns:
        np.ndarray of uint64: Indices that sort pos along dimension d.

    """
    cdef np.intp_t ndim = pos.shape[1]
    cdef int64_t l = 0
    cdef int64_t r = pos.shape[0]-1
    cdef uint64_t[:] idx
    idx = np.arange(pos.shape[0]).astype('uint64')
    cdef double *ptr_pos = NULL
    cdef uint64_t *ptr_idx = NULL
    if pos.shape[0] != 0:
        ptr_pos = &pos[0,0]
        ptr_idx = &idx[0]
    quickSort(ptr_pos, ptr_idx, ndim, d, l, r)
    return idx

def py_insertSort(np.ndarray[np.float64_t, ndim=2] pos, np.uint32_t d):
    r"""Get the indices required to sort coordinates along one dimension.

    Args:
        pos (np.ndarray of float64): (n,m) array of n m-D coordinates.
        d (np.uint32_t): Dimension that pos should be sorted along.

    Returns:
        np.ndarray of uint64: Indices that sort pos along dimension d.

    """
    cdef np.intp_t ndim = pos.shape[1]
    cdef int64_t l = 0
    cdef int64_t r = pos.shape[0]-1
    cdef uint64_t[:] idx
    idx = np.arange(pos.shape[0]).astype('uint64')
    cdef double *ptr_pos = NULL
    cdef uint64_t *ptr_idx = NULL
    if pos.shape[0] != 0:
        ptr_pos = &pos[0,0]
        ptr_idx = &idx[0]
    insertSort(ptr_pos, ptr_idx, ndim, d, l, r)
    return idx

def py_pivot(np.ndarray[np.float64_t, ndim=2] pos, np.uint32_t d):
    r"""Get the index of the median of medians along one dimension and indices
    that partition pos according to the median of medians.

    Args:
        pos (np.ndarray of float64): (n,m) array of n m-D coordinates.
        d (np.uint32_t): Dimension that pos should be partitioned along.

    Returns:
        tuple of int64 and np.ndarray of uint64: Index q of idx that is the
            pivot. All elements of idx before the pivot will be less than
            the pivot. If there is an odd number of points, the pivot will
            be the median.

    """
    cdef np.intp_t ndim = pos.shape[1]
    cdef int64_t l = 0
    cdef int64_t r = pos.shape[0]-1
    cdef uint64_t[:] idx
    idx = np.arange(pos.shape[0]).astype('uint64')
    cdef double *ptr_pos = NULL
    cdef uint64_t *ptr_idx = NULL
    if pos.shape[0] != 0:
        ptr_pos = &pos[0,0]
        ptr_idx = &idx[0]
    cdef int64_t q = pivot(ptr_pos, ptr_idx, ndim, d, l, r)
    return q, idx

def py_partition(np.ndarray[np.float64_t, ndim=2] pos, np.uint32_t d,
                 np.int64_t p):
    r"""Get the indices required to partition coordinates along one dimension.

    Args:
        pos (np.ndarray of float64): (n,m) array of n m-D coordinates.
        d (np.uint32_t): Dimension that pos should be partitioned along.
        p (np.int64_t): Element of pos[:,d] that should be used as the pivot
            to partition pos.

    Returns:
        tuple of int64 and np.ndarray of uint64: Location of the pivot in the
            partitioned array and the indices required to partition the array
            such that elements before the pivot are smaller and elements after
            the pivot are larger.

    """
    cdef np.intp_t ndim = pos.shape[1]
    cdef int64_t l = 0
    cdef int64_t r = pos.shape[0]-1
    cdef uint64_t[:] idx
    idx = np.arange(pos.shape[0]).astype('uint64')
    cdef double *ptr_pos = NULL
    cdef uint64_t *ptr_idx = NULL
    if pos.shape[0] != 0:
        ptr_pos = &pos[0,0]
        ptr_idx = &idx[0]
    cdef int64_t q = partition(ptr_pos, ptr_idx, ndim, d, l, r, p)
    return q, idx

def py_partition_given_pivot(np.ndarray[np.float64_t, ndim=2] pos,
                             np.uint32_t d, np.float64_t pval):
    r"""Get the indices required to partition coordinates along one dimension.

    Args:
        pos (np.ndarray of float64): (n,m) array of n m-D coordinates.
        d (np.uint32_t): Dimension that pos should be partitioned along.
        pval (np.float64_t): Value that should be used to partition pos.

    Returns:
        tuple of int64 and np.ndarray of uint64: Location of the largest value
            that is smaller than pval in partitioned array and the indices
            required to partition the array such that elements before the pivot
            are smaller and elements after the pivot are larger.

    """
    cdef np.intp_t ndim = pos.shape[1]
    cdef int64_t l = 0
    cdef int64_t r = pos.shape[0]-1
    cdef uint64_t[:] idx
    idx = np.arange(pos.shape[0]).astype('uint64')
    cdef double *ptr_pos = NULL
    cdef uint64_t *ptr_idx = NULL
    if pos.shape[0] != 0:
        ptr_pos = &pos[0,0]
        ptr_idx = &idx[0]
    cdef int64_t q = partition_given_pivot(ptr_pos, ptr_idx, ndim, d, l, r,
                                           pval)
    return q, idx

def py_select(np.ndarray[np.float64_t, ndim=2] pos, np.uint32_t d,
              np.int64_t t):
    r"""Get the indices required to partition coordinates such that the first
    t elements in pos[:,d] are the smallest t elements in pos[:,d].

    Args:
        pos (np.ndarray of float64): (n,m) array of n m-D coordinates.
        d (np.uint32_t): Dimension that pos should be partitioned along.
        t (np.int64_t): Number of smallest elements in pos[:,d] that should be
            partitioned.

    Returns:
        tuple of int64 and np.ndarray of uint64: Location of element t in the
            partitioned array and the indices required to partition the array
            such that elements before element t are smaller and elements after
            the pivot are larger.

    """
    cdef np.intp_t ndim = pos.shape[1]
    cdef int64_t l = 0
    cdef int64_t r = pos.shape[0]-1
    cdef uint64_t[:] idx
    idx = np.arange(pos.shape[0]).astype('uint64')
    cdef double *ptr_pos = NULL
    cdef uint64_t *ptr_idx = NULL
    if pos.shape[0] != 0:
        ptr_pos = &pos[0,0]
        ptr_idx = &idx[0]
    cdef int64_t q = select(ptr_pos, ptr_idx, ndim, d, l, r, t)
    return q, idx


def py_split(np.ndarray[np.float64_t, ndim=2] pos,
             np.ndarray[np.float64_t, ndim=1] mins = None,
             np.ndarray[np.float64_t, ndim=1] maxs = None,
             bool use_sliding_midpoint = False):
    r"""Get the indices required to split the positions equally along the
    largest dimension.

    Args:
        pos (np.ndarray of float64): (n,m) array of n m-D coordinates.
        mins (np.ndarray of float64, optional): (m,) array of mins. Defaults
            to None and is set to mins of pos along each dimension.
        maxs (np.ndarray of float64, optional): (m,) array of maxs. Defaults
            to None and is set to maxs of pos along each dimension.
        use_sliding_midpoint (bool, optional): If True, the sliding midpoint
             rule is used to split the positions. Defaults to False.

    Returns:
        tuple(int64, uint32, np.ndarray of uint64): The index of the split in
            the partitioned array, the dimension of the split, and the indices
            required to partition the array.

    """
    cdef np.intp_t npts = pos.shape[0]
    cdef np.intp_t ndim = pos.shape[1]
    cdef uint64_t Lidx = 0
    cdef uint64_t[:] idx
    idx = np.arange(pos.shape[0]).astype('uint64')
    cdef double *ptr_pos = NULL
    cdef uint64_t *ptr_idx = NULL
    cdef double *ptr_mins = NULL
    cdef double *ptr_maxs = NULL
    if (npts != 0) and (ndim != 0):
        if mins is None:
            mins = np.min(pos, axis=0)
        if maxs is None:
            maxs = np.max(pos, axis=0)
        ptr_pos = &pos[0,0]
        ptr_idx = &idx[0]
        ptr_mins = &mins[0]
        ptr_maxs = &maxs[0]
    cdef int64_t q = 0
    cdef double split_val = 0.0
    cdef cbool c_midpoint_flag = <cbool>use_sliding_midpoint
    cdef uint32_t dsplit = split(ptr_pos, ptr_idx, Lidx, npts, ndim,
                                 ptr_mins, ptr_maxs, q, split_val,
                                 c_midpoint_flag)
    return q, dsplit, idx



import numpy as np

from yt._typing import KnownFieldsT
from yt.fields.field_info_container import FieldInfoContainer
from yt.fields.particle_fields import add_union_field
from yt.frontends.enzo_e.misc import (
    get_listed_subparam,
    get_particle_mass_correction,
    nested_dict_get,
)

rho_units = "code_mass / code_length**3"
vel_units = "code_velocity"
acc_units = "code_velocity / code_time"
energy_units = "code_velocity**2"
b_units = "code_magnetic"

NODAL_FLAGS = {
    "bfieldi_x": [1, 0, 0],
    "bfieldi_y": [0, 1, 0],
    "bfieldi_z": [0, 0, 1],
}


class EnzoEFieldInfo(FieldInfoContainer):
    known_other_fields: KnownFieldsT = (
        ("velocity_x", (vel_units, ["velocity_x"], None)),
        ("velocity_y", (vel_units, ["velocity_y"], None)),
        ("velocity_z", (vel_units, ["velocity_z"], None)),
        ("acceleration_x", (acc_units, ["acceleration_x"], None)),
        ("acceleration_y", (acc_units, ["acceleration_y"], None)),
        ("acceleration_z", (acc_units, ["acceleration_z"], None)),
        ("density", (rho_units, ["density"], None)),
        ("density_total", (rho_units, ["total_density"], None)),
        ("total_energy", (energy_units, ["specific_total_energy"], None)),
        ("internal_energy", (energy_units, ["specific_thermal_energy"], None)),
        ("bfield_x", (b_units, [], None)),
        ("bfield_y", (b_units, [], None)),
        ("bfield_z", (b_units, [], None)),
        ("bfieldi_x", (b_units, [], None)),
        ("bfieldi_y", (b_units, [], None)),
        ("bfieldi_z", (b_units, [], None)),
    )

    known_particle_fields: KnownFieldsT = (
        ("x", ("code_length", ["particle_position_x"], None)),
        ("y", ("code_length", ["particle_position_y"], None)),
        ("z", ("code_length", ["particle_position_z"], None)),
        ("vx", (vel_units, ["particle_velocity_x"], None)),
        ("vy", (vel_units, ["particle_velocity_y"], None)),
        ("vz", (vel_units, ["particle_velocity_z"], None)),
        ("ax", (acc_units, ["particle_acceleration_x"], None)),
        ("ay", (acc_units, ["particle_acceleration_y"], None)),
        ("az", (acc_units, ["particle_acceleration_z"], None)),
        ("mass", ("code_mass", ["particle_mass"], None)),
    )

    def __init__(self, ds, field_list, slice_info=None):
        super().__init__(ds, field_list, slice_info=slice_info)

        # setup nodal flag information
        for field, arr in NODAL_FLAGS.items():
            if ("enzoe", field) in self:
                finfo = self["enzoe", field]
                finfo.nodal_flag = np.array(arr)

    def setup_fluid_fields(self):
        from yt.fields.magnetic_field import setup_magnetic_field_aliases

        self.setup_energy_field()
        setup_magnetic_field_aliases(self, "enzoe", [f"bfield_{ax}" for ax in "xyz"])

    def setup_energy_field(self):
        unit_system = self.ds.unit_system
        # check if we have a field for internal energy
        has_ie_field = ("enzoe", "internal_energy") in self.field_list
        # check if we need to account for magnetic energy
        vlct_params = get_listed_subparam(self.ds.parameters, "Method", "mhd_vlct", {})
        has_magnetic = "no_bfield" != vlct_params.get("mhd_choice", "no_bfield")

        # define the ("gas", "specific_thermal_energy") field
        # - this is already done for us if the simulation used the dual-energy
        #   formalism AND ("enzoe", "internal_energy") was saved to disk
        if not (self.ds.parameters["uses_dual_energy"] and has_ie_field):

            def _tot_minus_kin(field, data):
                return (
                    data["enzoe", "total_energy"]
                    - 0.5 * data["gas", "velocity_magnitude"] ** 2.0
                )

            if not has_magnetic:
                # thermal energy = total energy - kinetic energy
                self.add_field(
                    ("gas", "specific_thermal_energy"),
                    sampling_type="cell",
                    function=_tot_minus_kin,
                    units=unit_system["specific_energy"],
                )
            else:
                # thermal energy = total energy - kinetic energy - magnetic energy
                def _sub_b(field, data):
                    return (
                        _tot_minus_kin(field, data)
                        - data["gas", "magnetic_energy_density"]
                        / data["gas", "density"]
                    )

                self.add_field(
                    ("gas", "specific_thermal_energy"),
                    sampling_type="cell",
                    function=_sub_b,
                    units=unit_system["specific_energy"],
                )

    def setup_particle_fields(self, ptype, ftype="gas", num_neighbors=64):
        super().setup_particle_fields(ptype, ftype=ftype, num_neighbors=num_neighbors)
        self.setup_particle_mass_field(ptype)

    def setup_particle_mass_field(self, ptype):
        fname = "particle_mass"
        if ptype in self.ds.particle_unions:
            add_union_field(self, ptype, fname, "code_mass")
            return

        pdict = self.ds.parameters.get("Particle", None)
        if pdict is None:
            return

        constants = nested_dict_get(pdict, (ptype, "constants"), default=())
        if not constants:
            return

        # constants should be a tuple consisting of multiple tuples of (name, type, value).
        # When there is only one entry, the enclosing tuple gets stripped, so we put it back.
        if not isinstance(constants[0], tuple):
            constants = (constants,)
        names = [c[0] for c in constants]

        if "mass" not in names:
            return

        val = constants[names.index("mass")][2] * self.ds.mass_unit
        if not self.ds.index.io._particle_mass_is_mass:
            val = val * get_particle_mass_correction(self.ds)

        def _pmass(field, data):
            return val * data[ptype, "particle_ones"]

        self.add_field(
            (ptype, fname),
            function=_pmass,
            units="code_mass",
            sampling_type="particle",
        )



import os
from glob import glob

import numpy as np

from yt.utilities.grid_data_format.conversion.conversion_abc import Converter
from yt.utilities.on_demand_imports import _h5py as h5py

translation_dict = {}
translation_dict["density"] = "density"
translation_dict["total_energy"] = "specific_energy"
translation_dict["velocity_x"] = "velocity_x"
translation_dict["velocity_y"] = "velocity_y"
translation_dict["velocity_z"] = "velocity_z"
translation_dict["cell_centered_B_x"] = "mag_field_x"
translation_dict["cell_centered_B_y"] = "mag_field_y"
translation_dict["cell_centered_B_z"] = "mag_field_z"


class AthenaDistributedConverter(Converter):
    def __init__(self, basename, outname=None, source_dir=None, field_conversions=None):
        self.fields = []
        self.current_time = 0.0
        name = basename.split(".")
        self.ddn = int(name[1])
        if source_dir is None:
            source_dir = "./"
        self.source_dir = source_dir
        self.basename = name[0]
        if outname is None:
            outname = self.basename + ".%04i" % self.ddn + ".gdf"
        self.outname = outname
        if field_conversions is None:
            field_conversions = {}
        self.field_conversions = field_conversions
        self.handle = None

    def parse_line(self, line, grid):
        # grid is a dictionary
        splitup = line.strip().split()
        if "vtk" in splitup:
            grid["vtk_version"] = splitup[-1]
        elif "Really" in splitup:
            grid["time"] = splitup[-1]
            self.current_time = grid["time"]
        elif "PRIMITIVE" in splitup:
            grid["time"] = float(splitup[4].rstrip(","))
            grid["level"] = int(splitup[6].rstrip(","))
            grid["domain"] = int(splitup[8].rstrip(","))
            self.current_time = grid["time"]
        elif "DIMENSIONS" in splitup:
            grid["dimensions"] = np.array(splitup[-3:]).astype("int")
        elif "ORIGIN" in splitup:
            grid["left_edge"] = np.array(splitup[-3:]).astype("float64")
        elif "SPACING" in splitup:
            grid["dds"] = np.array(splitup[-3:]).astype("float64")
        elif "CELL_DATA" in splitup:
            grid["ncells"] = int(splitup[-1])
        elif "SCALARS" in splitup:
            field = splitup[1]
            grid["read_field"] = field
            grid["read_type"] = "scalar"
        elif "VECTORS" in splitup:
            field = splitup[1]
            grid["read_field"] = field
            grid["read_type"] = "vector"

    def write_gdf_field(self, fn, grid_number, field, data):
        f = self.handle
        ## --------- Store Grid Data --------- ##
        if "grid_%010i" % grid_number not in f["data"].keys():
            g = f["data"].create_group("grid_%010i" % grid_number)
        else:
            g = f["data"]["grid_%010i" % grid_number]
        name = field
        try:
            name = translation_dict[name]
        except Exception:
            pass
        # print 'Writing %s' % name
        if name not in g.keys():
            g.create_dataset(name, data=data)

    def read_and_write_index(self, basename, ddn, gdf_name):
        """Read Athena legacy vtk file from multiple cpus"""
        proc_names = glob(os.path.join(self.source_dir, "id*"))
        # print('Reading a dataset from %i Processor Files' % len(proc_names))
        N = len(proc_names)
        grid_dims = np.empty([N, 3], dtype="int64")
        grid_left_edges = np.empty([N, 3], dtype="float64")
        grid_dds = np.empty([N, 3], dtype="float64")
        grid_levels = np.zeros(N, dtype="int64")
        grid_parent_ids = -1 * np.ones(N, dtype="int64")
        grid_particle_counts = np.zeros([N, 1], dtype="int64")

        for i in range(N):
            if i == 0:
                fn = os.path.join(
                    self.source_dir, f"id{i}", basename + f".{ddn:04d}.vtk"
                )
            else:
                fn = os.path.join(
                    self.source_dir, f"id{i}", basename + f"-id{i}.{ddn:04d}.vtk"
                )

            print(f"Reading file {fn}")
            f = open(fn, "rb")
            grid = {}
            grid["read_field"] = None
            grid["read_type"] = None
            line = f.readline()
            while grid["read_field"] is None:
                self.parse_line(line, grid)
                if "SCALAR" in line.strip().split():
                    break
                if "VECTOR" in line.strip().split():
                    break
                if "TABLE" in line.strip().split():
                    break
                if len(line) == 0:
                    break
                del line
                line = f.readline()

            if len(line) == 0:
                break

            if np.prod(grid["dimensions"]) != grid["ncells"]:
                grid["dimensions"] -= 1
                grid["dimensions"][grid["dimensions"] == 0] = 1
            if np.prod(grid["dimensions"]) != grid["ncells"]:
                print(
                    "product of dimensions %i not equal to number of cells %i"
                    % (np.prod(grid["dimensions"]), grid["ncells"])
                )
                raise TypeError

            # Append all hierarchy info before reading this grid's data
            grid_dims[i] = grid["dimensions"]
            grid_left_edges[i] = grid["left_edge"]
            grid_dds[i] = grid["dds"]
            # grid_ncells[i]=grid['ncells']
            del grid

            f.close()
            del f
        f = self.handle

        ## --------- Begin level nodes --------- ##
        g = f.create_group("gridded_data_format")
        g.attrs["format_version"] = np.float32(1.0)
        g.attrs["data_software"] = "athena"
        f.create_group("data")
        f.create_group("field_types")
        f.create_group("particle_types")
        pars_g = f.create_group("simulation_parameters")

        gles = grid_left_edges
        gdims = grid_dims
        dle = np.min(gles, axis=0)
        dre = np.max(gles + grid_dims * grid_dds, axis=0)
        glis = ((gles - dle) / grid_dds).astype("int64")

        ddims = (dre - dle) / grid_dds[0]

        # grid_left_index
        f.create_dataset("grid_left_index", data=glis)
        # grid_dimensions
        f.create_dataset("grid_dimensions", data=gdims)

        # grid_level
        f.create_dataset("grid_level", data=grid_levels)

        ## ----------QUESTIONABLE NEXT LINE--------- ##
        # This data needs two dimensions for now.
        f.create_dataset("grid_particle_count", data=grid_particle_counts)

        # grid_parent_id
        f.create_dataset("grid_parent_id", data=grid_parent_ids)

        ## --------- Done with top level nodes --------- ##

        pars_g.attrs["refine_by"] = np.int64(1)
        pars_g.attrs["dimensionality"] = np.int64(3)
        pars_g.attrs["domain_dimensions"] = ddims
        pars_g.attrs["current_time"] = self.current_time
        pars_g.attrs["domain_left_edge"] = dle
        pars_g.attrs["domain_right_edge"] = dre
        pars_g.attrs["unique_identifier"] = "athenatest"
        pars_g.attrs["cosmological_simulation"] = np.int64(0)
        pars_g.attrs["num_ghost_zones"] = np.int64(0)
        pars_g.attrs["field_ordering"] = np.int64(1)
        pars_g.attrs["boundary_conditions"] = np.int64([0] * 6)  # For Now

        # Extra pars:
        # pars_g.attrs['n_cells'] = grid['ncells']
        pars_g.attrs["vtk_version"] = 1.0

        # Add particle types
        # Nothing to do here

        # Add particle field attributes
        # f.close()

    def read_and_write_data(self, basename, ddn, gdf_name):
        proc_names = glob(os.path.join(self.source_dir, "id*"))
        # print('Reading a dataset from %i Processor Files' % len(proc_names))
        N = len(proc_names)
        for i in range(N):
            if i == 0:
                fn = os.path.join(
                    self.source_dir, f"id{i}", basename + f".{ddn:04d}.vtk"
                )
            else:
                fn = os.path.join(
                    self.source_dir, +f"id{i}", basename + f"-id{i}.{ddn:04d}.vtk"
                )
            f = open(fn, "rb")
            # print('Reading data from %s' % fn)
            line = f.readline()
            while line != "":
                if len(line) == 0:
                    break
                splitup = line.strip().split()

                if "DIMENSIONS" in splitup:
                    grid_dims = np.array(splitup[-3:]).astype("int")
                    line = f.readline()
                    continue
                elif "CELL_DATA" in splitup:
                    grid_ncells = int(splitup[-1])
                    line = f.readline()
                    if np.prod(grid_dims) != grid_ncells:
                        grid_dims -= 1
                        grid_dims[grid_dims == 0] = 1
                    if np.prod(grid_dims) != grid_ncells:
                        print(
                            "product of dimensions %i not equal to number of cells %i"
                            % (np.prod(grid_dims), grid_ncells)
                        )
                        raise TypeError
                    break
                else:
                    del line
                    line = f.readline()
            read_table = False
            while line != "":
                if len(line) == 0:
                    break
                splitup = line.strip().split()
                if "SCALARS" in splitup:
                    field = splitup[1]
                    if not read_table:
                        line = f.readline()  # Read the lookup table line
                        read_table = True
                    data = np.fromfile(f, dtype=">f4", count=grid_ncells).reshape(
                        grid_dims, order="F"
                    )
                    if i == 0:
                        self.fields.append(field)
                    # print('writing field %s' % field)
                    self.write_gdf_field(gdf_name, i, field, data)
                    read_table = False

                elif "VECTORS" in splitup:
                    field = splitup[1]
                    data = np.fromfile(f, dtype=">f4", count=3 * grid_ncells)
                    data_x = data[0::3].reshape(grid_dims, order="F")
                    data_y = data[1::3].reshape(grid_dims, order="F")
                    data_z = data[2::3].reshape(grid_dims, order="F")
                    if i == 0:
                        self.fields.append(field + "_x")
                        self.fields.append(field + "_y")
                        self.fields.append(field + "_z")

                    # print('writing field %s' % field)
                    self.write_gdf_field(gdf_name, i, field + "_x", data_x)
                    self.write_gdf_field(gdf_name, i, field + "_y", data_y)
                    self.write_gdf_field(gdf_name, i, field + "_z", data_z)
                    del data, data_x, data_y, data_z
                del line
                line = f.readline()
            f.close()
            del f

        f = self.handle
        field_g = f["field_types"]
        # Add Field Attributes
        for name in self.fields:
            tname = name
            try:
                tname = translation_dict[name]
            except Exception:
                pass
            this_field = field_g.create_group(tname)
            if name in self.field_conversions.keys():
                this_field.attrs["field_to_cgs"] = self.field_conversions[name]
            else:
                this_field.attrs["field_to_cgs"] = np.float64("1.0")  # For Now

    def convert(self, index=True, data=True):
        self.handle = h5py.File(self.outname, mode="a")
        if index:
            self.read_and_write_index(self.basename, self.ddn, self.outname)
        if data:
            self.read_and_write_data(self.basename, self.ddn, self.outname)
        self.handle.close()


class AthenaConverter(Converter):
    def __init__(self, basename, outname=None, field_conversions=None):
        self.fields = []
        self.basename = basename
        name = basename.split(".")
        fn = "%s.%04i" % (name[0], int(name[1]))
        self.ddn = int(name[1])
        self.basename = fn
        if outname is None:
            outname = fn + ".gdf"
        self.outname = outname
        if field_conversions is None:
            field_conversions = {}
        self.field_conversions = field_conversions

    def parse_line(self, line, grid):
        # grid is a dictionary
        splitup = line.strip().split()
        if "vtk" in splitup:
            grid["vtk_version"] = splitup[-1]
        elif "Really" in splitup:
            grid["time"] = splitup[-1]
        elif "DIMENSIONS" in splitup:
            grid["dimensions"] = np.array(splitup[-3:]).astype("int")
        elif "ORIGIN" in splitup:
            grid["left_edge"] = np.array(splitup[-3:]).astype("float64")
        elif "SPACING" in splitup:
            grid["dds"] = np.array(splitup[-3:]).astype("float64")
        elif "CELL_DATA" in splitup:
            grid["ncells"] = int(splitup[-1])
        elif "SCALARS" in splitup:
            field = splitup[1]
            grid["read_field"] = field
            grid["read_type"] = "scalar"
        elif "VECTORS" in splitup:
            field = splitup[1]
            grid["read_field"] = field
            grid["read_type"] = "vector"

    def read_grid(self, filename):
        """Read Athena legacy vtk file from single cpu"""
        f = open(filename, "rb")
        # print('Reading from %s'%filename)
        grid = {}
        grid["read_field"] = None
        grid["read_type"] = None
        table_read = False
        line = f.readline()
        while line != "":
            while grid["read_field"] is None:
                self.parse_line(line, grid)
                if grid["read_type"] == "vector":
                    break
                if not table_read:
                    line = f.readline()
                if "TABLE" in line.strip().split():
                    table_read = True
                if len(line) == 0:
                    break

            if len(line) == 0:
                break
            if np.prod(grid["dimensions"]) != grid["ncells"]:
                grid["dimensions"] -= 1
            if np.prod(grid["dimensions"]) != grid["ncells"]:
                print(
                    "product of dimensions %i not equal to number of cells %i"
                    % (np.prod(grid["dimensions"]), grid["ncells"])
                )
                raise TypeError

            if grid["read_type"] == "scalar":
                grid[grid["read_field"]] = np.fromfile(
                    f, dtype=">f4", count=grid["ncells"]
                ).reshape(grid["dimensions"], order="F")
                self.fields.append(grid["read_field"])
            elif grid["read_type"] == "vector":
                data = np.fromfile(f, dtype=">f4", count=3 * grid["ncells"])
                grid[grid["read_field"] + "_x"] = data[0::3].reshape(
                    grid["dimensions"], order="F"
                )
                grid[grid["read_field"] + "_y"] = data[1::3].reshape(
                    grid["dimensions"], order="F"
                )
                grid[grid["read_field"] + "_z"] = data[2::3].reshape(
                    grid["dimensions"], order="F"
                )
                self.fields.append(grid["read_field"] + "_x")
                self.fields.append(grid["read_field"] + "_y")
                self.fields.append(grid["read_field"] + "_z")
            else:
                raise TypeError
            grid["read_field"] = None
            grid["read_type"] = None
            line = f.readline()
            if len(line) == 0:
                break
        grid["right_edge"] = grid["left_edge"] + grid["dds"] * (grid["dimensions"])
        return grid

    def write_to_gdf(self, fn, grid):
        f = h5py.File(fn, mode="a")

        ## --------- Begin level nodes --------- ##
        g = f.create_group("gridded_data_format")
        g.attrs["format_version"] = np.float32(1.0)
        g.attrs["data_software"] = "athena"
        data_g = f.create_group("data")
        field_g = f.create_group("field_types")
        f.create_group("particle_types")
        pars_g = f.create_group("simulation_parameters")

        dle = grid["left_edge"]  # True only in this case of one grid for the domain
        gles = np.array([grid["left_edge"]])
        gdims = np.array([grid["dimensions"]])
        glis = ((gles - dle) / grid["dds"]).astype("int64")

        # grid_left_index
        f.create_dataset("grid_left_index", data=glis)
        # grid_dimensions
        f.create_dataset("grid_dimensions", data=gdims)

        levels = np.array([0]).astype("int64")  # unigrid example
        # grid_level
        f.create_dataset("grid_level", data=levels)

        ## ----------QUESTIONABLE NEXT LINE--------- ##
        # This data needs two dimensions for now.
        n_particles = np.array([[0]]).astype("int64")
        # grid_particle_count
        f.create_dataset("grid_particle_count", data=n_particles)

        # Assume -1 means no parent.
        parent_ids = np.array([-1]).astype("int64")
        # grid_parent_id
        f.create_dataset("grid_parent_id", data=parent_ids)

        ## --------- Done with top level nodes --------- ##

        f.create_group("index")

        ## --------- Store Grid Data --------- ##

        g0 = data_g.create_group("grid_%010i" % 0)
        for field in self.fields:
            name = field
            if field in translation_dict.keys():
                name = translation_dict[name]
            if name not in g0.keys():
                g0.create_dataset(name, data=grid[field])

        ## --------- Store Particle Data --------- ##

        # Nothing to do

        ## --------- Attribute Tables --------- ##

        pars_g.attrs["refine_by"] = np.int64(1)
        pars_g.attrs["dimensionality"] = np.int64(3)
        pars_g.attrs["domain_dimensions"] = grid["dimensions"]
        try:
            pars_g.attrs["current_time"] = grid["time"]
        except Exception:
            pars_g.attrs["current_time"] = 0.0
        pars_g.attrs["domain_left_edge"] = grid["left_edge"]  # For Now
        pars_g.attrs["domain_right_edge"] = grid["right_edge"]  # For Now
        pars_g.attrs["unique_identifier"] = "athenatest"
        pars_g.attrs["cosmological_simulation"] = np.int64(0)
        pars_g.attrs["num_ghost_zones"] = np.int64(0)
        pars_g.attrs["field_ordering"] = np.int64(0)
        pars_g.attrs["boundary_conditions"] = np.int64([0] * 6)  # For Now

        # Extra pars:
        pars_g.attrs["n_cells"] = grid["ncells"]
        pars_g.attrs["vtk_version"] = grid["vtk_version"]

        # Add Field Attributes
        for name in g0.keys():
            tname = name
            try:
                tname = translation_dict[name]
            except Exception:
                pass
            this_field = field_g.create_group(tname)
        if name in self.field_conversions.keys():
            this_field.attrs["field_to_cgs"] = self.field_conversions[name]
        else:
            this_field.attrs["field_to_cgs"] = np.float64("1.0")  # For Now

        # Add particle types
        # Nothing to do here

        # Add particle field attributes
        f.close()

    def convert(self):
        grid = self.read_grid(self.basename + ".vtk")
        self.write_to_gdf(self.outname, grid)


# import sys
# if __name__ == '__main__':
#     n = sys.argv[-1]
#     n = n.split('.')
#     fn = '%s.%04i'%(n[0],int(n[1]))
#     grid = read_grid(fn+'.vtk')
#     write_to_hdf5(fn+'.gdf',grid)



import numpy as np
from numpy.testing import assert_almost_equal, assert_equal

from yt import SlicePlot
from yt.frontends.stream.data_structures import hexahedral_connectivity
from yt.loaders import load_hexahedral_mesh

# Field information


def test_stream_hexahedral():
    np.random.seed(0x4D3D3D3)
    Nx, Ny, Nz = 32, 18, 24
    # Note what we're doing here -- we are creating a randomly spaced mesh, but
    # because of how the accumulate operation works, we also reset the leftmost
    # cell boundary to 0.0.
    cell_x = np.random.random(Nx + 1)
    cell_x /= cell_x.sum()
    cell_x = np.add.accumulate(cell_x)
    cell_x[0] = 0.0

    cell_y = np.random.random(Ny + 1)
    cell_y /= cell_y.sum()
    cell_y = np.add.accumulate(cell_y)
    cell_y[0] = 0.0

    cell_z = np.random.random(Nz + 1)
    cell_z /= cell_z.sum()
    cell_z = np.add.accumulate(cell_z)
    cell_z[0] = 0.0

    coords, conn = hexahedral_connectivity(cell_x, cell_y, cell_z)
    data = {"random_field": np.random.random((Nx, Ny, Nz))}
    bbox = np.array([[0.0, 1.0], [0.0, 1.0], [0.0, 1.0]])
    ds = load_hexahedral_mesh(data, conn, coords, bbox=bbox)
    dd = ds.all_data()
    # raise RuntimeError
    assert_almost_equal(float(dd[("gas", "cell_volume")].sum(dtype="float64")), 1.0)
    assert_equal(dd[("index", "ones")].size, Nx * Ny * Nz)
    # Now we try it with a standard mesh
    cell_x = np.linspace(0.0, 1.0, Nx + 1)
    cell_y = np.linspace(0.0, 1.0, Ny + 1)
    cell_z = np.linspace(0.0, 1.0, Nz + 1)
    coords, conn = hexahedral_connectivity(cell_x, cell_y, cell_z)
    data = {"random_field": np.random.random((Nx, Ny, Nz))}
    bbox = np.array([[0.0, 1.0], [0.0, 1.0], [0.0, 1.0]])
    ds = load_hexahedral_mesh(data, conn, coords, bbox=bbox)
    dd = ds.all_data()
    assert_almost_equal(float(dd[("gas", "cell_volume")].sum(dtype="float64")), 1.0)
    assert_equal(dd[("index", "ones")].size, Nx * Ny * Nz)
    assert_almost_equal(dd[("index", "dx")].to_ndarray(), 1.0 / Nx)
    assert_almost_equal(dd[("index", "dy")].to_ndarray(), 1.0 / Ny)
    assert_almost_equal(dd[("index", "dz")].to_ndarray(), 1.0 / Nz)

    s = SlicePlot(ds, "x", "random_field")
    s.render()
    s.frb[("stream", "random_field")]



import os
import re
import string
import time
import weakref
from collections import defaultdict
from functools import cached_property

import numpy as np
from more_itertools import always_iterable

from yt.data_objects.index_subobjects.grid_patch import AMRGridPatch
from yt.data_objects.static_output import Dataset
from yt.fields.field_info_container import NullFunc
from yt.frontends.enzo.misc import cosmology_get_units
from yt.funcs import get_pbar, iter_fields, setdefaultattr
from yt.geometry.geometry_handler import YTDataChunk
from yt.geometry.grid_geometry_handler import GridIndex
from yt.utilities.logger import ytLogger as mylog
from yt.utilities.on_demand_imports import _h5py as h5py, _libconf as libconf

from .fields import EnzoFieldInfo


class EnzoGrid(AMRGridPatch):
    """
    Class representing a single Enzo Grid instance.
    """

    def __init__(self, id, index):
        """
        Returns an instance of EnzoGrid with *id*, associated with
        *filename* and *index*.
        """
        # All of the field parameters will be passed to us as needed.
        AMRGridPatch.__init__(self, id, filename=None, index=index)
        self._children_ids = []
        self._parent_id = -1
        self.Level = -1

    def set_filename(self, filename):
        """
        Intelligently set the filename.
        """
        if filename is None:
            self.filename = filename
            return
        if self.index._strip_path:
            self.filename = os.path.join(
                self.index.directory, os.path.basename(filename)
            )
        elif filename[0] == os.path.sep:
            self.filename = filename
        else:
            self.filename = os.path.join(self.index.directory, filename)
        return

    @property
    def Parent(self):
        if self._parent_id == -1:
            return None
        return self.index.grids[self._parent_id - self._id_offset]

    @property
    def Children(self):
        return [self.index.grids[cid - self._id_offset] for cid in self._children_ids]

    @property
    def NumberOfActiveParticles(self):
        if not hasattr(self.index, "grid_active_particle_count"):
            return {}
        id = self.id - self._id_offset
        nap = {
            ptype: self.index.grid_active_particle_count[ptype][id]
            for ptype in self.index.grid_active_particle_count
        }
        return nap


class EnzoGridInMemory(EnzoGrid):
    __slots__ = ["proc_num"]

    def set_filename(self, filename):
        pass


class EnzoGridGZ(EnzoGrid):
    __slots__ = ()

    def retrieve_ghost_zones(self, n_zones, fields, all_levels=False, smoothed=False):
        NGZ = self.ds.parameters.get("NumberOfGhostZones", 3)
        if n_zones > NGZ:
            return EnzoGrid.retrieve_ghost_zones(
                self, n_zones, fields, all_levels, smoothed
            )

        # ----- Below is mostly the original code, except we remove the field
        # ----- access section
        # We will attempt this by creating a datacube that is exactly bigger
        # than the grid by nZones*dx in each direction
        nl = self.get_global_startindex() - n_zones
        new_left_edge = nl * self.dds + self.ds.domain_left_edge
        # Something different needs to be done for the root grid, though
        level = self.Level
        kwargs = {
            "dims": self.ActiveDimensions + 2 * n_zones,
            "num_ghost_zones": n_zones,
            "use_pbar": False,
        }
        # This should update the arguments to set the field parameters to be
        # those of this grid.
        kwargs.update(self.field_parameters)
        if smoothed:
            # cube = self.index.smoothed_covering_grid(
            #    level, new_left_edge, new_right_edge, **kwargs)
            cube = self.index.smoothed_covering_grid(level, new_left_edge, **kwargs)
        else:
            cube = self.index.covering_grid(level, new_left_edge, **kwargs)
        # ----- This is EnzoGrid.get_data, duplicated here mostly for
        # ----  efficiency's sake.
        start_zone = NGZ - n_zones
        if start_zone == 0:
            end_zone = None
        else:
            end_zone = -(NGZ - n_zones)
        sl = tuple(slice(start_zone, end_zone) for i in range(3))
        if fields is None:
            return cube
        for field in iter_fields(fields):
            if field in self.field_list:
                conv_factor = 1.0
                if field in self.ds.field_info:
                    conv_factor = self.ds.field_info[field]._convert_function(self)
                if self.ds.field_info[field].sampling_type == "particle":
                    continue
                temp = self.index.io._read_raw_data_set(self, field)
                temp = temp.swapaxes(0, 2)
                cube.field_data[field] = np.multiply(temp, conv_factor, temp)[sl]
        return cube


class EnzoHierarchy(GridIndex):
    _strip_path = False
    grid = EnzoGrid
    _preload_implemented = True

    def __init__(self, ds, dataset_type):
        self.dataset_type = dataset_type
        self.index_filename = os.path.abspath(f"{ds.parameter_filename}.hierarchy")
        if os.path.getsize(self.index_filename) == 0:
            raise OSError(-1, "File empty", self.index_filename)
        self.directory = os.path.dirname(self.index_filename)

        # For some reason, r8 seems to want Float64
        if "CompilerPrecision" in ds and ds["CompilerPrecision"] == "r4":
            self.float_type = "float32"
        else:
            self.float_type = "float64"

        GridIndex.__init__(self, ds, dataset_type)
        # sync it back
        self.dataset.dataset_type = self.dataset_type

    def _count_grids(self):
        self.num_grids = None
        test_grid = test_grid_id = None
        self.num_stars = 0
        for line in rlines(open(self.index_filename, "rb")):
            if (
                line.startswith("BaryonFileName")
                or line.startswith("ParticleFileName")
                or line.startswith("FileName ")
            ):
                test_grid = line.split("=")[-1].strip().rstrip()
            if line.startswith("NumberOfStarParticles"):
                self.num_stars = int(line.split("=")[-1])
            if line.startswith("Grid "):
                if self.num_grids is None:
                    self.num_grids = int(line.split("=")[-1])
                test_grid_id = int(line.split("=")[-1])
                if test_grid is not None:
                    break
        self._guess_dataset_type(self.ds.dimensionality, test_grid, test_grid_id)

    def _guess_dataset_type(self, rank, test_grid, test_grid_id):
        if test_grid[0] != os.path.sep:
            test_grid = os.path.join(self.directory, test_grid)
        if not os.path.exists(test_grid):
            test_grid = os.path.join(self.directory, os.path.basename(test_grid))
            mylog.debug("Your data uses the annoying hardcoded path.")
            self._strip_path = True
        if self.dataset_type is not None:
            return
        if rank == 3:
            mylog.debug("Detected packed HDF5")
            if self.parameters.get("WriteGhostZones", 0) == 1:
                self.dataset_type = "enzo_packed_3d_gz"
                self.grid = EnzoGridGZ
            else:
                self.dataset_type = "enzo_packed_3d"
        elif rank == 2:
            mylog.debug("Detect packed 2D")
            self.dataset_type = "enzo_packed_2d"
        elif rank == 1:
            mylog.debug("Detect packed 1D")
            self.dataset_type = "enzo_packed_1d"
        else:
            raise NotImplementedError

    # Sets are sorted, so that won't work!
    def _parse_index(self):
        def _next_token_line(token, f):
            for line in f:
                if line.startswith(token):
                    return line.split()[2:]

        pattern = r"Pointer: Grid\[(\d*)\]->NextGrid(Next|This)Level = (\d*)\s+$"
        patt = re.compile(pattern)
        f = open(self.index_filename)
        self.grids = [self.grid(1, self)]
        self.grids[0].Level = 0
        si, ei, LE, RE, fn, npart = [], [], [], [], [], []
        pbar = get_pbar("Parsing Hierarchy ", self.num_grids)
        version = self.dataset.parameters.get("VersionNumber", None)
        params = self.dataset.parameters
        if version is None and "Internal" in params:
            version = float(params["Internal"]["Provenance"]["VersionNumber"])
        if version >= 3.0:
            active_particles = True
            nap = {
                ap_type: []
                for ap_type in params["Physics"]["ActiveParticles"][
                    "ActiveParticlesEnabled"
                ]
            }
        else:
            if "AppendActiveParticleType" in self.parameters:
                nap = {}
                active_particles = True
                for type in self.parameters.get("AppendActiveParticleType", []):
                    nap[type] = []
            else:
                nap = None
                active_particles = False
        for grid_id in range(self.num_grids):
            pbar.update(grid_id + 1)
            # We will unroll this list
            si.append(_next_token_line("GridStartIndex", f))
            ei.append(_next_token_line("GridEndIndex", f))
            LE.append(_next_token_line("GridLeftEdge", f))
            RE.append(_next_token_line("GridRightEdge", f))
            nb = int(_next_token_line("NumberOfBaryonFields", f)[0])
            fn.append([None])
            if nb > 0:
                fn[-1] = _next_token_line("BaryonFileName", f)
            npart.append(int(_next_token_line("NumberOfParticles", f)[0]))
            # Below we find out what active particles exist in this grid,
            # and add their counts individually.
            if active_particles:
                ptypes = _next_token_line("PresentParticleTypes", f)
                counts = [int(c) for c in _next_token_line("ParticleTypeCounts", f)]
                for ptype in self.parameters.get("AppendActiveParticleType", []):
                    if ptype in ptypes:
                        nap[ptype].append(counts[ptypes.index(ptype)])
                    else:
                        nap[ptype].append(0)
            if nb == 0 and npart[-1] > 0:
                fn[-1] = _next_token_line("ParticleFileName", f)
            for line in f:
                if len(line) < 2:
                    break
                if line.startswith("Pointer:"):
                    vv = patt.findall(line)[0]
                    self.__pointer_handler(vv)
        pbar.finish()
        self._fill_arrays(ei, si, LE, RE, npart, nap)
        temp_grids = np.empty(self.num_grids, dtype="object")
        temp_grids[:] = self.grids
        self.grids = temp_grids
        self.filenames = fn

    def _initialize_grid_arrays(self):
        super()._initialize_grid_arrays()
        if "AppendActiveParticleType" in self.parameters.keys() and len(
            self.parameters["AppendActiveParticleType"]
        ):
            gac = {
                ptype: np.zeros(self.num_grids, dtype="i4")
                for ptype in self.parameters["AppendActiveParticleType"]
            }
            self.grid_active_particle_count = gac

    def _fill_arrays(self, ei, si, LE, RE, npart, nap):
        self.grid_dimensions.flat[:] = ei
        self.grid_dimensions -= np.array(si, dtype="i4")
        self.grid_dimensions += 1
        self.grid_left_edge.flat[:] = LE
        self.grid_right_edge.flat[:] = RE
        self.grid_particle_count.flat[:] = npart
        if nap is not None:
            for ptype in nap:
                self.grid_active_particle_count[ptype].flat[:] = nap[ptype]

    def __pointer_handler(self, m):
        sgi = int(m[2]) - 1
        if sgi == -1:
            return  # if it's 0, then we're done with that lineage
        # Okay, so, we have a pointer.  We make a new grid, with an id of the length+1
        # (recall, Enzo grids are 1-indexed)
        self.grids.append(self.grid(len(self.grids) + 1, self))
        # We'll just go ahead and make a weakref to cache
        second_grid = self.grids[sgi]  # zero-indexed already
        first_grid = self.grids[int(m[0]) - 1]
        if m[1] == "Next":
            first_grid._children_ids.append(second_grid.id)
            second_grid._parent_id = first_grid.id
            second_grid.Level = first_grid.Level + 1
        elif m[1] == "This":
            if first_grid.Parent is not None:
                first_grid.Parent._children_ids.append(second_grid.id)
                second_grid._parent_id = first_grid._parent_id
            second_grid.Level = first_grid.Level
        self.grid_levels[sgi] = second_grid.Level

    def _rebuild_top_grids(self, level=0):
        mylog.info("Rebuilding grids on level %s", level)
        cmask = self.grid_levels.flat == (level + 1)
        cmsum = cmask.sum()
        mask = np.zeros(self.num_grids, dtype="bool")
        for grid in self.select_grids(level):
            mask[:] = 0
            LE = self.grid_left_edge[grid.id - grid._id_offset]
            RE = self.grid_right_edge[grid.id - grid._id_offset]
            grids, grid_i = self.get_box_grids(LE, RE)
            mask[grid_i] = 1
            grid._children_ids = []
            cgrids = self.grids[(mask * cmask).astype("bool")]
            mylog.info("%s: %s / %s", grid, len(cgrids), cmsum)
            for cgrid in cgrids:
                grid._children_ids.append(cgrid.id)
                cgrid._parent_id = grid.id
        mylog.info("Finished rebuilding")

    def _populate_grid_objects(self):
        for g, f in zip(self.grids, self.filenames):
            g._prepare_grid()
            g._setup_dx()
            g.set_filename(f[0])
        del self.filenames  # No longer needed.
        self.max_level = self.grid_levels.max()

    def _detect_active_particle_fields(self):
        ap_list = self.dataset["AppendActiveParticleType"]
        _fields = {ap: [] for ap in ap_list}
        fields = []
        for ptype in self.dataset["AppendActiveParticleType"]:
            select_grids = self.grid_active_particle_count[ptype].flat
            if not np.any(select_grids):
                current_ptypes = self.dataset.particle_types
                new_ptypes = [p for p in current_ptypes if p != ptype]
                self.dataset.particle_types = new_ptypes
                self.dataset.particle_types_raw = new_ptypes
                continue
            if ptype != "DarkMatter":
                gs = self.grids[select_grids > 0]
                g = gs[0]
                handle = h5py.File(g.filename, "r")
                grid_group = handle[f"/Grid{g.id:08d}"]
                for pname in ["Active Particles", "Particles"]:
                    if pname in grid_group:
                        break
                else:
                    raise RuntimeError("Could not find active particle group in data.")
                node = grid_group[pname]
                for ptype in (str(p) for p in node):
                    if ptype not in _fields:
                        continue
                    for field in (str(f) for f in node[ptype]):
                        _fields[ptype].append(field)
                        if node[ptype][field].ndim > 1:
                            self.io._array_fields[field] = (
                                node[ptype][field].shape[1:],
                            )
                    fields += [(ptype, field) for field in _fields.pop(ptype)]
                handle.close()
        return set(fields)

    def _setup_derived_fields(self):
        super()._setup_derived_fields()
        aps = self.dataset.parameters.get("AppendActiveParticleType", [])
        for fname, field in self.ds.field_info.items():
            if not field.sampling_type == "particle":
                continue
            if isinstance(fname, tuple):
                continue
            if field._function is NullFunc:
                continue
            for apt in aps:
                dd = field._copy_def()
                dd.pop("name")
                self.ds.field_info.add_field((apt, fname), sampling_type="cell", **dd)

    def _detect_output_fields(self):
        self.field_list = []
        # Do this only on the root processor to save disk work.
        if self.comm.rank in (0, None):
            mylog.info("Gathering a field list (this may take a moment.)")
            field_list = set()
            random_sample = self._generate_random_grids()
            for grid in random_sample:
                if not hasattr(grid, "filename"):
                    continue
                try:
                    gf = self.io._read_field_names(grid)
                except self.io._read_exception as e:
                    raise OSError("Grid %s is a bit funky?", grid.id) from e
                mylog.debug("Grid %s has: %s", grid.id, gf)
                field_list = field_list.union(gf)
            if "AppendActiveParticleType" in self.dataset.parameters:
                ap_fields = self._detect_active_particle_fields()
                field_list = list(set(field_list).union(ap_fields))
                if not any(f[0] == "io" for f in field_list):
                    if "io" in self.dataset.particle_types_raw:
                        ptypes_raw = list(self.dataset.particle_types_raw)
                        ptypes_raw.remove("io")
                        self.dataset.particle_types_raw = tuple(ptypes_raw)

                    if "io" in self.dataset.particle_types:
                        ptypes = list(self.dataset.particle_types)
                        ptypes.remove("io")
                        self.dataset.particle_types = tuple(ptypes)
            ptypes = self.dataset.particle_types
            ptypes_raw = self.dataset.particle_types_raw
        else:
            field_list = None
            ptypes = None
            ptypes_raw = None
        self.field_list = list(self.comm.mpi_bcast(field_list))
        self.dataset.particle_types = list(self.comm.mpi_bcast(ptypes))
        self.dataset.particle_types_raw = list(self.comm.mpi_bcast(ptypes_raw))

    def _generate_random_grids(self):
        if self.num_grids > 40:
            starter = np.random.randint(0, 20)
            random_sample = np.mgrid[starter : len(self.grids) - 1 : 20j].astype(
                "int32"
            )
            # We also add in a bit to make sure that some of the grids have
            # particles
            gwp = self.grid_particle_count > 0
            if np.any(gwp) and not np.any(gwp[(random_sample,)]):
                # We just add one grid.  This is not terribly efficient.
                first_grid = np.where(gwp)[0][0]
                random_sample.resize((21,))
                random_sample[-1] = first_grid
                mylog.debug("Added additional grid %s", first_grid)
            mylog.debug("Checking grids: %s", random_sample.tolist())
        else:
            random_sample = np.mgrid[0 : max(len(self.grids), 1)].astype("int32")
        return self.grids[(random_sample,)]

    def _get_particle_type_counts(self):
        try:
            ret = {}
            for ptype in self.grid_active_particle_count:
                ret[ptype] = self.grid_active_particle_count[ptype].sum()
            return ret
        except AttributeError:
            return super()._get_particle_type_counts()

    def find_particles_by_type(self, ptype, max_num=None, additional_fields=None):
        """
        Returns a structure of arrays with all of the particles'
        positions, velocities, masses, types, IDs, and attributes for
        a particle type **ptype** for a maximum of **max_num**
        particles.  If non-default particle fields are used, provide
        them in **additional_fields**.
        """
        # Not sure whether this routine should be in the general HierarchyType.
        if self.grid_particle_count.sum() == 0:
            mylog.info("Data contains no particles.")
            return None
        if additional_fields is None:
            additional_fields = [
                "metallicity_fraction",
                "creation_time",
                "dynamical_time",
            ]
        pfields = [f for f in self.field_list if f.startswith("particle_")]
        nattr = self.dataset["NumberOfParticleAttributes"]
        if nattr > 0:
            pfields += additional_fields[:nattr]
        # Find where the particles reside and count them
        if max_num is None:
            max_num = 1e100
        total = 0
        pstore = []
        for level in range(self.max_level, -1, -1):
            for grid in self.select_grids(level):
                index = np.where(grid["particle_type"] == ptype)[0]
                total += len(index)
                pstore.append(index)
                if total >= max_num:
                    break
            if total >= max_num:
                break
        result = None
        if total > 0:
            result = {}
            for p in pfields:
                result[p] = np.zeros(total, "float64")
            # Now we retrieve data for each field
            ig = count = 0
            for level in range(self.max_level, -1, -1):
                for grid in self.select_grids(level):
                    nidx = len(pstore[ig])
                    if nidx > 0:
                        for p in pfields:
                            result[p][count : count + nidx] = grid[p][pstore[ig]]
                        count += nidx
                    ig += 1
                    if count >= total:
                        break
                if count >= total:
                    break
            # Crop data if retrieved more than max_num
            if count > max_num:
                for p in pfields:
                    result[p] = result[p][0:max_num]
        return result


class EnzoHierarchyInMemory(EnzoHierarchy):
    grid = EnzoGridInMemory

    @cached_property
    def enzo(self):
        import enzo

        return enzo

    def __init__(self, ds, dataset_type=None):
        self.dataset_type = dataset_type
        self.float_type = "float64"
        self.dataset = weakref.proxy(ds)  # for _obtain_enzo
        self.float_type = self.enzo.hierarchy_information["GridLeftEdge"].dtype
        self.directory = os.getcwd()
        GridIndex.__init__(self, ds, dataset_type)

    def _initialize_data_storage(self):
        pass

    def _count_grids(self):
        self.num_grids = self.enzo.hierarchy_information["GridDimensions"].shape[0]

    def _parse_index(self):
        self._copy_index_structure()
        mylog.debug("Copying reverse tree")
        reverse_tree = self.enzo.hierarchy_information["GridParentIDs"].ravel().tolist()
        # Initial setup:
        mylog.debug("Reconstructing parent-child relationships")
        grids = []
        # We enumerate, so it's 0-indexed id and 1-indexed pid
        self.filenames = ["-1"] * self.num_grids
        for id, pid in enumerate(reverse_tree):
            grids.append(self.grid(id + 1, self))
            grids[-1].Level = self.grid_levels[id, 0]
            if pid > 0:
                grids[-1]._parent_id = pid
                grids[pid - 1]._children_ids.append(grids[-1].id)
        self.max_level = self.grid_levels.max()
        mylog.debug("Preparing grids")
        self.grids = np.empty(len(grids), dtype="object")
        for i, grid in enumerate(grids):
            if (i % 1e4) == 0:
                mylog.debug("Prepared % 7i / % 7i grids", i, self.num_grids)
            grid.filename = "Inline_processor_%07i" % (self.grid_procs[i, 0])
            grid._prepare_grid()
            grid._setup_dx()
            grid.proc_num = self.grid_procs[i, 0]
            self.grids[i] = grid
        mylog.debug("Prepared")

    def _initialize_grid_arrays(self):
        EnzoHierarchy._initialize_grid_arrays(self)
        self.grid_procs = np.zeros((self.num_grids, 1), "int32")

    def _copy_index_structure(self):
        # Dimensions are important!
        self.grid_dimensions[:] = self.enzo.hierarchy_information["GridEndIndices"][:]
        self.grid_dimensions -= self.enzo.hierarchy_information["GridStartIndices"][:]
        self.grid_dimensions += 1
        self.grid_left_edge[:] = self.enzo.hierarchy_information["GridLeftEdge"][:]
        self.grid_right_edge[:] = self.enzo.hierarchy_information["GridRightEdge"][:]
        self.grid_levels[:] = self.enzo.hierarchy_information["GridLevels"][:]
        self.grid_procs = self.enzo.hierarchy_information["GridProcs"].copy()
        self.grid_particle_count[:] = self.enzo.hierarchy_information[
            "GridNumberOfParticles"
        ][:]

    def save_data(self, *args, **kwargs):
        pass

    _cached_field_list = None
    _cached_derived_field_list = None

    def _generate_random_grids(self):
        my_rank = self.comm.rank
        my_grids = self.grids[self.grid_procs.ravel() == my_rank]
        if len(my_grids) > 40:
            starter = np.random.randint(0, 20)
            random_sample = np.mgrid[starter : len(my_grids) - 1 : 20j].astype("int32")
            mylog.debug("Checking grids: %s", random_sample.tolist())
        else:
            random_sample = np.mgrid[0 : max(len(my_grids) - 1, 1)].astype("int32")
        return my_grids[(random_sample,)]

    def _chunk_io(self, dobj, cache=True, local_only=False):
        gfiles = defaultdict(list)
        gobjs = getattr(dobj._current_chunk, "objs", dobj._chunk_info)
        for g in gobjs:
            gfiles[g.filename].append(g)
        for fn in sorted(gfiles):
            if local_only:
                gobjs = [g for g in gfiles[fn] if g.proc_num == self.comm.rank]
                gfiles[fn] = gobjs
            gs = gfiles[fn]
            count = self._count_selection(dobj, gs)
            yield YTDataChunk(dobj, "io", gs, count, cache=cache)


class EnzoHierarchy1D(EnzoHierarchy):
    def _fill_arrays(self, ei, si, LE, RE, npart, nap):
        self.grid_dimensions[:, :1] = ei
        self.grid_dimensions[:, :1] -= np.array(si, dtype="i4")
        self.grid_dimensions += 1
        self.grid_left_edge[:, :1] = LE
        self.grid_right_edge[:, :1] = RE
        self.grid_particle_count.flat[:] = npart
        self.grid_left_edge[:, 1:] = 0.0
        self.grid_right_edge[:, 1:] = 1.0
        self.grid_dimensions[:, 1:] = 1
        if nap is not None:
            raise NotImplementedError


class EnzoHierarchy2D(EnzoHierarchy):
    def _fill_arrays(self, ei, si, LE, RE, npart, nap):
        self.grid_dimensions[:, :2] = ei
        self.grid_dimensions[:, :2] -= np.array(si, dtype="i4")
        self.grid_dimensions += 1
        self.grid_left_edge[:, :2] = LE
        self.grid_right_edge[:, :2] = RE
        self.grid_particle_count.flat[:] = npart
        self.grid_left_edge[:, 2] = 0.0
        self.grid_right_edge[:, 2] = 1.0
        self.grid_dimensions[:, 2] = 1
        if nap is not None:
            raise NotImplementedError


class EnzoDataset(Dataset):
    """
    Enzo-specific output, set at a fixed time.
    """

    _index_class = EnzoHierarchy
    _field_info_class = EnzoFieldInfo

    def __init__(
        self,
        filename,
        dataset_type=None,
        parameter_override=None,
        conversion_override=None,
        storage_filename=None,
        units_override=None,
        unit_system="cgs",
        default_species_fields=None,
    ):
        """
        This class is a stripped down class that simply reads and parses
        *filename* without looking at the index.  *dataset_type* gets passed
        to the index to pre-determine the style of data-output.  However,
        it is not strictly necessary.  Optionally you may specify a
        *parameter_override* dictionary that will override anything in the
        parameter file and a *conversion_override* dictionary that consists
        of {fieldname : conversion_to_cgs} that will override the #DataCGS.
        """
        self.fluid_types += ("enzo",)
        if filename.endswith(".hierarchy"):
            filename = filename[:-10]
        if parameter_override is None:
            parameter_override = {}
        self._parameter_override = parameter_override
        if conversion_override is None:
            conversion_override = {}
        self._conversion_override = conversion_override
        self.storage_filename = storage_filename
        Dataset.__init__(
            self,
            filename,
            dataset_type,
            units_override=units_override,
            unit_system=unit_system,
            default_species_fields=default_species_fields,
        )

    def _setup_1d(self):
        self._index_class = EnzoHierarchy1D
        self.domain_left_edge = np.concatenate([[self.domain_left_edge], [0.0, 0.0]])
        self.domain_right_edge = np.concatenate([[self.domain_right_edge], [1.0, 1.0]])

    def _setup_2d(self):
        self._index_class = EnzoHierarchy2D
        self.domain_left_edge = np.concatenate([self.domain_left_edge, [0.0]])
        self.domain_right_edge = np.concatenate([self.domain_right_edge, [1.0]])

    def get_parameter(self, parameter, type=None):
        """
        Gets a parameter not in the parameterDict.
        """
        if parameter in self.parameters:
            return self.parameters[parameter]
        for line in open(self.parameter_filename):
            if line.find("#") >= 1:  # Keep the commented lines
                line = line[: line.find("#")]
            line = line.strip().rstrip()
            if len(line) < 2:
                continue
            try:
                param, vals = map(string.strip, map(string.rstrip, line.split("=")))
            except ValueError:
                mylog.error("ValueError: '%s'", line)
            if parameter == param:
                if type is None:
                    t = vals.split()
                else:
                    t = map(type, vals.split())
                if len(t) == 1:
                    self.parameters[param] = t[0]
                else:
                    self.parameters[param] = t
                if param.endswith("Units") and not param.startswith("Temperature"):
                    dataType = param[:-5]
                    self.conversion_factors[dataType] = self.parameters[param]
                return self.parameters[parameter]

        return ""

    @cached_property
    def unique_identifier(self) -> str:
        if "CurrentTimeIdentifier" in self.parameters:
            # enzo2
            return str(self.parameters["CurrentTimeIdentifier"])
        elif "MetaDataDatasetUUID" in self.parameters:
            # enzo2
            return str(self.parameters["MetaDataDatasetUUID"])
        elif "Internal" in self.parameters:
            # enzo3
            return str(
                self.parameters["Internal"]["Provenance"]["CurrentTimeIdentidier"]
            )
        else:
            return super().unique_identifier

    def _parse_parameter_file(self):
        """
        Parses the parameter file and establishes the various
        dictionaries.
        """
        # Let's read the file
        with open(self.parameter_filename) as f:
            line = f.readline().strip()
            f.seek(0)
            if line == "Internal:":
                self._parse_enzo3_parameter_file(f)
            else:
                self._parse_enzo2_parameter_file(f)

    def _parse_enzo3_parameter_file(self, f):
        self.parameters = p = libconf.load(f)
        sim = p["SimulationControl"]
        internal = p["Internal"]
        phys = p["Physics"]
        self.refine_by = sim["AMR"]["RefineBy"]
        self._periodicity = tuple(
            a == 3 for a in sim["Domain"]["LeftFaceBoundaryCondition"]
        )
        self.dimensionality = sim["Domain"]["TopGridRank"]
        self.domain_dimensions = np.array(
            sim["Domain"]["TopGridDimensions"], dtype="int64"
        )
        self.domain_left_edge = np.array(
            sim["Domain"]["DomainLeftEdge"], dtype="float64"
        )
        self.domain_right_edge = np.array(
            sim["Domain"]["DomainRightEdge"], dtype="float64"
        )
        self.gamma = phys["Hydro"]["Gamma"]
        self.current_time = internal["InitialTime"]
        self.cosmological_simulation = phys["Cosmology"]["ComovingCoordinates"]
        if self.cosmological_simulation == 1:
            cosmo = phys["Cosmology"]
            self.current_redshift = internal["CosmologyCurrentRedshift"]
            self.omega_lambda = cosmo["OmegaLambdaNow"]
            self.omega_matter = cosmo["OmegaMatterNow"]
            self.hubble_constant = cosmo["HubbleConstantNow"]
        else:
            self.current_redshift = 0.0
            self.omega_lambda = 0.0
            self.omega_matter = 0.0
            self.hubble_constant = 0.0
            self.cosmological_simulation = 0
        self.particle_types = ["DarkMatter"] + phys["ActiveParticles"][
            "ActiveParticlesEnabled"
        ]
        self.particle_types = tuple(self.particle_types)
        self.particle_types_raw = self.particle_types
        if self.dimensionality == 1:
            self._setup_1d()
        elif self.dimensionality == 2:
            self._setup_2d()

    def _parse_enzo2_parameter_file(self, f):
        for line in (l.strip() for l in f):
            if (len(line) < 2) or ("=" not in line):
                continue
            param, vals = (i.strip() for i in line.split("=", 1))
            # First we try to decipher what type of value it is.
            vals = vals.split()
            # Special case approaching.
            if "(do" in vals:
                vals = vals[:1]
            if len(vals) == 0:
                pcast = str  # Assume NULL output
            else:
                v = vals[0]
                # Figure out if it's castable to floating point:
                try:
                    float(v)
                except ValueError:
                    pcast = str
                else:
                    if any("." in v or "e+" in v or "e-" in v for v in vals):
                        pcast = float
                    elif v == "inf":
                        pcast = str
                    else:
                        pcast = int
            # Now we figure out what to do with it.
            if len(vals) == 0:
                vals = ""
            elif len(vals) == 1:
                vals = pcast(vals[0])
            else:
                vals = np.array([pcast(i) for i in vals if i != "-99999"])
            if param.startswith("Append"):
                if param not in self.parameters:
                    self.parameters[param] = []
                self.parameters[param].append(vals)
            else:
                self.parameters[param] = vals
        self.refine_by = self.parameters["RefineBy"]
        _periodicity = tuple(
            always_iterable(self.parameters["LeftFaceBoundaryCondition"] == 3)
        )
        self.dimensionality = self.parameters["TopGridRank"]

        if self.dimensionality > 1:
            self.domain_dimensions = self.parameters["TopGridDimensions"]
            if len(self.domain_dimensions) < 3:
                tmp = self.domain_dimensions.tolist()
                tmp.append(1)
                self.domain_dimensions = np.array(tmp)
                _periodicity += (False,)
            self.domain_left_edge = np.array(
                self.parameters["DomainLeftEdge"], "float64"
            ).copy()
            self.domain_right_edge = np.array(
                self.parameters["DomainRightEdge"], "float64"
            ).copy()
        else:
            self.domain_left_edge = np.array(
                self.parameters["DomainLeftEdge"], "float64"
            )
            self.domain_right_edge = np.array(
                self.parameters["DomainRightEdge"], "float64"
            )
            self.domain_dimensions = np.array(
                [self.parameters["TopGridDimensions"], 1, 1]
            )
            _periodicity += (False, False)
        assert len(_periodicity) == 3
        self._periodicity = _periodicity

        self.gamma = self.parameters["Gamma"]
        if self.parameters["ComovingCoordinates"]:
            self.cosmological_simulation = 1
            self.current_redshift = self.parameters["CosmologyCurrentRedshift"]
            self.omega_lambda = self.parameters["CosmologyOmegaLambdaNow"]
            self.omega_matter = self.parameters["CosmologyOmegaMatterNow"]
            self.omega_radiation = self.parameters.get(
                "CosmologyOmegaRadiationNow", 0.0
            )
            self.hubble_constant = self.parameters["CosmologyHubbleConstantNow"]
        else:
            self.current_redshift = 0.0
            self.omega_lambda = 0.0
            self.omega_matter = 0.0
            self.hubble_constant = 0.0
            self.cosmological_simulation = 0
        self.particle_types = []
        self.current_time = self.parameters["InitialTime"]
        if (
            self.parameters["NumberOfParticles"] > 0
            and "AppendActiveParticleType" in self.parameters.keys()
        ):
            # If this is the case, then we know we should have a DarkMatter
            # particle type, and we don't need the "io" type.
            self.parameters["AppendActiveParticleType"].append("DarkMatter")
        else:
            # We do not have an "io" type for Enzo particles if the
            # ActiveParticle machinery is on, as we simply will ignore any of
            # the non-DarkMatter particles in that case.  However, for older
            # datasets, we call this particle type "io".
            self.particle_types = ["io"]
        for ptype in self.parameters.get("AppendActiveParticleType", []):
            self.particle_types.append(ptype)
        self.particle_types = tuple(self.particle_types)
        self.particle_types_raw = self.particle_types

        if self.dimensionality == 1:
            self._setup_1d()
        elif self.dimensionality == 2:
            self._setup_2d()

    def _set_code_unit_attributes(self):
        if self.cosmological_simulation:
            k = cosmology_get_units(
                self.hubble_constant,
                self.omega_matter,
                self.parameters["CosmologyComovingBoxSize"],
                self.parameters["CosmologyInitialRedshift"],
                self.current_redshift,
            )
            # Now some CGS values
            box_size = self.parameters["CosmologyComovingBoxSize"]
            setdefaultattr(self, "length_unit", self.quan(box_size, "Mpccm/h"))
            setdefaultattr(
                self,
                "mass_unit",
                self.quan(k["urho"], "g/cm**3") * (self.length_unit.in_cgs()) ** 3,
            )
            setdefaultattr(self, "time_unit", self.quan(k["utim"], "s"))
            setdefaultattr(self, "velocity_unit", self.quan(k["uvel"], "cm/s"))
        else:
            if "LengthUnits" in self.parameters:
                length_unit = self.parameters["LengthUnits"]
                mass_unit = self.parameters["DensityUnits"] * length_unit**3
                time_unit = self.parameters["TimeUnits"]
            elif "SimulationControl" in self.parameters:
                units = self.parameters["SimulationControl"]["Units"]
                length_unit = units["Length"]
                mass_unit = units["Density"] * length_unit**3
                time_unit = units["Time"]
            else:
                mylog.warning("Setting 1.0 in code units to be 1.0 cm")
                mylog.warning("Setting 1.0 in code units to be 1.0 s")
                length_unit = mass_unit = time_unit = 1.0

            setdefaultattr(self, "length_unit", self.quan(length_unit, "cm"))
            setdefaultattr(self, "mass_unit", self.quan(mass_unit, "g"))
            setdefaultattr(self, "time_unit", self.quan(time_unit, "s"))
            setdefaultattr(self, "velocity_unit", self.length_unit / self.time_unit)

        density_unit = self.mass_unit / self.length_unit**3
        magnetic_unit = np.sqrt(4 * np.pi * density_unit) * self.velocity_unit
        magnetic_unit = np.float64(magnetic_unit.in_cgs())
        setdefaultattr(self, "magnetic_unit", self.quan(magnetic_unit, "gauss"))

    @classmethod
    def _is_valid(cls, filename, *args, **kwargs):
        if str(filename).endswith(".hierarchy"):
            return True
        return os.path.exists(f"{filename}.hierarchy")

    @classmethod
    def _guess_candidates(cls, base, directories, files):
        candidates = [
            _
            for _ in files
            if _.endswith(".hierarchy")
            and os.path.exists(os.path.join(base, _.rsplit(".", 1)[0]))
        ]
        # Typically, Enzo won't have nested outputs.
        return candidates, (len(candidates) == 0)


class EnzoDatasetInMemory(EnzoDataset):
    _index_class = EnzoHierarchyInMemory
    _dataset_type = "enzo_inline"

    def __init__(self, parameter_override=None, conversion_override=None):
        self.fluid_types += ("enzo",)
        if parameter_override is None:
            parameter_override = {}
        self._parameter_override = parameter_override
        if conversion_override is None:
            conversion_override = {}
        self._conversion_override = conversion_override

        Dataset.__init__(self, "InMemoryParameterFile", self._dataset_type)

    def _parse_parameter_file(self):
        enzo = self._obtain_enzo()
        self._input_filename = "cycle%08i" % (
            enzo.yt_parameter_file["NumberOfPythonCalls"]
        )
        self.parameters["CurrentTimeIdentifier"] = time.time()
        self.parameters.update(enzo.yt_parameter_file)
        self.conversion_factors.update(enzo.conversion_factors)
        for i in self.parameters:
            if isinstance(self.parameters[i], tuple):
                self.parameters[i] = np.array(self.parameters[i])
            if i.endswith("Units") and not i.startswith("Temperature"):
                dataType = i[:-5]
                self.conversion_factors[dataType] = self.parameters[i]
        self.domain_left_edge = self.parameters["DomainLeftEdge"].copy()
        self.domain_right_edge = self.parameters["DomainRightEdge"].copy()
        for i in self.conversion_factors:
            if isinstance(self.conversion_factors[i], tuple):
                self.conversion_factors[i] = np.array(self.conversion_factors[i])
        for p, v in self._parameter_override.items():
            self.parameters[p] = v
        for p, v in self._conversion_override.items():
            self.conversion_factors[p] = v
        self.refine_by = self.parameters["RefineBy"]
        self._periodicity = tuple(
            always_iterable(self.parameters["LeftFaceBoundaryCondition"] == 3)
        )
        self.dimensionality = self.parameters["TopGridRank"]
        self.domain_dimensions = self.parameters["TopGridDimensions"]
        self.current_time = self.parameters["InitialTime"]
        if self.parameters["ComovingCoordinates"]:
            self.cosmological_simulation = 1
            self.current_redshift = self.parameters["CosmologyCurrentRedshift"]
            self.omega_lambda = self.parameters["CosmologyOmegaLambdaNow"]
            self.omega_matter = self.parameters["CosmologyOmegaMatterNow"]
            self.hubble_constant = self.parameters["CosmologyHubbleConstantNow"]
        else:
            self.current_redshift = 0.0
            self.omega_lambda = 0.0
            self.omega_matter = 0.0
            self.hubble_constant = 0.0
            self.cosmological_simulation = 0

    def _obtain_enzo(self):
        import enzo

        return enzo

    @classmethod
    def _is_valid(cls, filename, *args, **kwargs):
        return False


# These next two functions are taken from
# http://www.reddit.com/r/Python/comments/6hj75/reverse_file_iterator/c03vms4
# Credit goes to "Brian" on Reddit


def rblocks(f, blocksize=4096):
    """Read file as series of blocks from end of file to start.

    The data itself is in normal order, only the order of the blocks is reversed.
    ie. "hello world" -> ["ld","wor", "lo ", "hel"]
    Note that the file must be opened in binary mode.
    """
    if "b" not in f.mode.lower():
        raise Exception("File must be opened using binary mode.")
    size = os.stat(f.name).st_size
    fullblocks, lastblock = divmod(size, blocksize)

    # The first(end of file) block will be short, since this leaves
    # the rest aligned on a blocksize boundary.  This may be more
    # efficient than having the last (first in file) block be short
    f.seek(-lastblock, 2)
    yield f.read(lastblock).decode("ascii")

    for i in range(fullblocks - 1, -1, -1):
        f.seek(i * blocksize)
        yield f.read(blocksize).decode("ascii")


def rlines(f, keepends=False):
    """Iterate through the lines of a file in reverse order.

    If keepends is true, line endings are kept as part of the line.
    """
    buf = ""
    for block in rblocks(f):
        buf = block + buf
        lines = buf.splitlines(keepends)
        # Return all lines except the first (since may be partial)
        if lines:
            lines.reverse()
            buf = lines.pop()  # Last line becomes end of new first line.
            yield from lines
    yield buf  # First line.



from contextlib import contextmanager
from functools import cached_property
from itertools import product, repeat
from typing import Tuple

import numpy as np
from unyt import unyt_array

import yt.geometry.particle_deposit as particle_deposit
import yt.geometry.particle_smooth as particle_smooth
from yt.data_objects.selection_objects.data_selection_objects import (
    YTSelectionContainer,
)
from yt.geometry.particle_oct_container import ParticleOctreeContainer
from yt.units.dimensions import length  # type: ignore
from yt.utilities.exceptions import (
    YTFieldTypeNotFound,
    YTInvalidPositionArray,
    YTParticleDepositionNotImplemented,
)
from yt.utilities.lib.geometry_utils import compute_morton
from yt.utilities.logger import ytLogger as mylog


def cell_count_cache(func):
    def cc_cache_func(self, dobj):
        if hash(dobj.selector) != self._last_selector_id:
            self._cell_count = -1
        rv = func(self, dobj)
        self._cell_count = rv.shape[0]
        self._last_selector_id = hash(dobj.selector)
        return rv

    return cc_cache_func


class OctreeSubset(YTSelectionContainer):
    _spatial = True
    _num_ghost_zones = 0
    _type_name = "octree_subset"
    _skip_add = True
    _con_args: Tuple[str, ...] = ("base_region", "domain", "ds")
    _domain_offset = 0
    _cell_count = -1
    _block_order = "C"

    def __init__(self, base_region, domain, ds, num_zones=2, num_ghost_zones=0):
        super().__init__(ds, None)
        self._num_zones = num_zones
        self._num_ghost_zones = num_ghost_zones
        self.domain = domain
        self.domain_id = domain.domain_id
        self.ds = domain.ds
        self._index = self.ds.index
        self.oct_handler = domain.oct_handler
        self._last_mask = None
        self._last_selector_id = None
        self.base_region = base_region
        self.base_selector = base_region.selector

    def __getitem__(self, key):
        tr = super().__getitem__(key)
        try:
            fields = self._determine_fields(key)
        except YTFieldTypeNotFound:
            return tr
        finfo = self.ds._get_field_info(fields[0])
        if not finfo.sampling_type == "particle":
            # We may need to reshape the field, if it is being queried from
            # field_data.  If it's already cached, it just passes through.
            if len(tr.shape) < 4:
                tr = self._reshape_vals(tr)
            return tr
        return tr

    @property
    def nz(self):
        return self._num_zones + 2 * self._num_ghost_zones

    def _reshape_vals(self, arr):
        nz = self.nz
        if len(arr.shape) <= 2:
            n_oct = arr.shape[0] // (nz**3)
        elif arr.shape[-1] == 3:
            n_oct = arr.shape[-2]
        else:
            n_oct = arr.shape[-1]
        if arr.size == nz * nz * nz * n_oct:
            new_shape = (nz, nz, nz, n_oct)
        elif arr.size == nz * nz * nz * n_oct * 3:
            new_shape = (nz, nz, nz, n_oct, 3)
        else:
            raise RuntimeError
        # Note that if arr is already F-contiguous, this *shouldn't* copy the
        # data.  But, it might.  However, I can't seem to figure out how to
        # make the assignment to .shape, which *won't* copy the data, make the
        # resultant array viewed in Fortran order.
        arr = arr.reshape(new_shape, order="F")
        return arr

    def mask_refinement(self, selector):
        mask = self.oct_handler.mask(selector, domain_id=self.domain_id)
        return mask

    def select_blocks(self, selector):
        mask = self.oct_handler.mask(selector, domain_id=self.domain_id)
        slicer = OctreeSubsetBlockSlice(self, self.ds)
        for i, sl in slicer:
            yield sl, np.atleast_3d(mask[i, ...])

    def select_tcoords(self, dobj):
        # These will not be pre-allocated, which can be a problem for speed and
        # memory usage.
        dts, ts = [], []
        for sl, mask in self.select_blocks(dobj.selector):
            sl.child_mask = np.asfortranarray(mask)
            dt, t = dobj.selector.get_dt(sl)
            dts.append(dt)
            ts.append(t)
        if len(dts) == len(ts) == 0:
            return np.empty(0, "f8"), np.empty(0, "f8")
        return np.concatenate(dts), np.concatenate(ts)

    @cached_property
    def domain_ind(self):
        return self.oct_handler.domain_ind(self.selector)

    def deposit(self, positions, fields=None, method=None, kernel_name="cubic"):
        r"""Operate on the mesh, in a particle-against-mesh fashion, with
        exclusively local input.

        This uses the octree indexing system to call a "deposition" operation
        (defined in yt/geometry/particle_deposit.pyx) that can take input from
        several particles (local to the mesh) and construct some value on the
        mesh.  The canonical example is to sum the total mass in a mesh cell
        and then divide by its volume.

        Parameters
        ----------
        positions : array_like (Nx3)
            The positions of all of the particles to be examined.  A new
            indexed octree will be constructed on these particles.
        fields : list of arrays
            All the necessary fields for computing the particle operation.  For
            instance, this might include mass, velocity, etc.
        method : string
            This is the "method name" which will be looked up in the
            `particle_deposit` namespace as `methodname_deposit`.  Current
            methods include `count`, `simple_smooth`, `sum`, `std`, `cic`,
            `weighted_mean`, `mesh_id`, and `nearest`.
        kernel_name : string, default 'cubic'
            This is the name of the smoothing kernel to use. Current supported
            kernel names include `cubic`, `quartic`, `quintic`, `wendland2`,
            `wendland4`, and `wendland6`.

        Returns
        -------
        List of fortran-ordered, mesh-like arrays.
        """
        # Here we perform our particle deposition.
        if fields is None:
            fields = []
        cls = getattr(particle_deposit, f"deposit_{method}", None)
        if cls is None:
            raise YTParticleDepositionNotImplemented(method)
        nz = self.nz
        nvals = (nz, nz, nz, (self.domain_ind >= 0).sum())
        if np.max(self.domain_ind) >= nvals[-1]:
            print(
                "nocts, domain_ind >= 0, max {} {} {}".format(
                    self.oct_handler.nocts, nvals[-1], np.max(self.domain_ind)
                )
            )
            raise Exception()
        # We allocate number of zones, not number of octs
        op = cls(nvals, kernel_name)
        op.initialize()
        mylog.debug(
            "Depositing %s (%s^3) particles into %s Octs",
            positions.shape[0],
            positions.shape[0] ** 0.3333333,
            nvals[-1],
        )
        positions.convert_to_units("code_length")
        pos = positions.d
        # We should not need the following if we know in advance all our fields
        # need no casting.
        fields = [np.ascontiguousarray(f, dtype="float64") for f in fields]
        op.process_octree(
            self.oct_handler,
            self.domain_ind,
            pos,
            fields,
            self.domain_id,
            self._domain_offset,
        )
        vals = op.finalize()
        if vals is None:
            return
        return np.asfortranarray(vals)

    def mesh_sampling_particle_field(self, positions, mesh_field, lvlmax=None):
        r"""Operate on the particles, in a mesh-against-particle
        fashion, with exclusively local input.

        This uses the octree indexing system to call a "mesh
        sampling" operation (defined in yt/geometry/particle_deposit.pyx).
        For each particle, the function returns the value of the cell
        containing the particle.

        Parameters
        ----------
        positions : array_like (Nx3)
            The positions of all of the particles to be examined.
        mesh_field : array_like (M,)
            The value of the field to deposit.
        lvlmax : array_like (N), optional
            If provided, the maximum level where to look for cells

        Returns
        -------
        List of fortran-ordered, particle-like arrays containing the
        value of the mesh at the location of the particles.
        """
        # Here we perform our particle deposition.
        npart = positions.shape[0]
        nocts = (self.domain_ind >= 0).sum()
        # We allocate number of zones, not number of octs
        op = particle_deposit.CellIdentifier(npart, "none")
        op.initialize(npart)
        mylog.debug(
            "Depositing %s Octs onto %s (%s^3) particles",
            nocts,
            positions.shape[0],
            positions.shape[0] ** 0.3333333,
        )
        pos = positions.to("code_length").value.astype("float64")

        op.process_octree(
            self.oct_handler,
            self.domain_ind,
            pos,
            None,
            self.domain_id,
            self._domain_offset,
            lvlmax=lvlmax,
        )

        igrid, icell = op.finalize()

        igrid = np.asfortranarray(igrid)
        icell = np.asfortranarray(icell)

        # Some particle may fall outside of the local domain, so we
        # need to be careful here
        ids = igrid * 8 + icell
        mask = ids >= 0

        # Fill the return array
        ret = np.empty(npart)
        ret[mask] = mesh_field[ids[mask]]
        ret[~mask] = np.nan

        return ret

    def smooth(
        self,
        positions,
        fields=None,
        index_fields=None,
        method=None,
        create_octree=False,
        nneighbors=64,
        kernel_name="cubic",
    ):
        r"""Operate on the mesh, in a particle-against-mesh fashion, with
        non-local input.

        This uses the octree indexing system to call a "smoothing" operation
        (defined in yt/geometry/particle_smooth.pyx) that can take input from
        several (non-local) particles and construct some value on the mesh.
        The canonical example is to conduct a smoothing kernel operation on the
        mesh.

        Parameters
        ----------
        positions : array_like (Nx3)
            The positions of all of the particles to be examined.  A new
            indexed octree will be constructed on these particles.
        fields : list of arrays
            All the necessary fields for computing the particle operation.  For
            instance, this might include mass, velocity, etc.
        index_fields : list of arrays
            All of the fields defined on the mesh that may be used as input to
            the operation.
        method : string
            This is the "method name" which will be looked up in the
            `particle_smooth` namespace as `methodname_smooth`.  Current
            methods include `volume_weighted`, `nearest`, `idw`,
            `nth_neighbor`, and `density`.
        create_octree : bool
            Should we construct a new octree for indexing the particles?  In
            cases where we are applying an operation on a subset of the
            particles used to construct the mesh octree, this will ensure that
            we are able to find and identify all relevant particles.
        nneighbors : int, default 64
            The number of neighbors to examine during the process.
        kernel_name : string, default 'cubic'
            This is the name of the smoothing kernel to use. Current supported
            kernel names include `cubic`, `quartic`, `quintic`, `wendland2`,
            `wendland4`, and `wendland6`.

        Returns
        -------
        List of fortran-ordered, mesh-like arrays.
        """
        # Here we perform our particle deposition.
        positions.convert_to_units("code_length")
        if create_octree:
            morton = compute_morton(
                positions[:, 0],
                positions[:, 1],
                positions[:, 2],
                self.ds.domain_left_edge,
                self.ds.domain_right_edge,
            )
            morton.sort()
            particle_octree = ParticleOctreeContainer(
                [1, 1, 1],
                self.ds.domain_left_edge,
                self.ds.domain_right_edge,
                num_zones=self._nz,
            )
            # This should ensure we get everything within one neighbor of home.
            particle_octree.n_ref = nneighbors * 2
            particle_octree.add(morton)
            particle_octree.finalize(self.domain_id)
            pdom_ind = particle_octree.domain_ind(self.selector)
        else:
            particle_octree = self.oct_handler
            pdom_ind = self.domain_ind
        if fields is None:
            fields = []
        if index_fields is None:
            index_fields = []
        cls = getattr(particle_smooth, f"{method}_smooth", None)
        if cls is None:
            raise YTParticleDepositionNotImplemented(method)
        nz = self.nz
        mdom_ind = self.domain_ind
        nvals = (nz, nz, nz, (mdom_ind >= 0).sum())
        op = cls(nvals, len(fields), nneighbors, kernel_name)
        op.initialize()
        mylog.debug(
            "Smoothing %s particles into %s Octs", positions.shape[0], nvals[-1]
        )
        # Pointer operations within 'process_octree' require arrays to be
        # contiguous cf. https://bitbucket.org/yt_analysis/yt/issues/1079
        fields = [np.ascontiguousarray(f, dtype="float64") for f in fields]
        op.process_octree(
            self.oct_handler,
            mdom_ind,
            positions,
            self.fcoords,
            fields,
            self.domain_id,
            self._domain_offset,
            self.ds.periodicity,
            index_fields,
            particle_octree,
            pdom_ind,
            self.ds.geometry,
        )
        # If there are 0s in the smoothing field this will not throw an error,
        # but silently return nans for vals where dividing by 0
        # Same as what is currently occurring, but suppressing the div by zero
        # error.
        with np.errstate(invalid="ignore"):
            vals = op.finalize()
        if isinstance(vals, list):
            vals = [np.asfortranarray(v) for v in vals]
        else:
            vals = np.asfortranarray(vals)
        return vals

    def particle_operation(
        self, positions, fields=None, method=None, nneighbors=64, kernel_name="cubic"
    ):
        r"""Operate on particles, in a particle-against-particle fashion.

        This uses the octree indexing system to call a "smoothing" operation
        (defined in yt/geometry/particle_smooth.pyx) that expects to be called
        in a particle-by-particle fashion.  For instance, the canonical example
        of this would be to compute the Nth nearest neighbor, or to compute the
        density for a given particle based on some kernel operation.

        Many of the arguments to this are identical to those used in the smooth
        and deposit functions.  Note that the `fields` argument must not be
        empty, as these fields will be modified in place.

        Parameters
        ----------
        positions : array_like (Nx3)
            The positions of all of the particles to be examined.  A new
            indexed octree will be constructed on these particles.
        fields : list of arrays
            All the necessary fields for computing the particle operation.  For
            instance, this might include mass, velocity, etc.  One of these
            will likely be modified in place.
        method : string
            This is the "method name" which will be looked up in the
            `particle_smooth` namespace as `methodname_smooth`.
        nneighbors : int, default 64
            The number of neighbors to examine during the process.
        kernel_name : string, default 'cubic'
            This is the name of the smoothing kernel to use. Current supported
            kernel names include `cubic`, `quartic`, `quintic`, `wendland2`,
            `wendland4`, and `wendland6`.

        Returns
        -------
        Nothing.

        """
        # Here we perform our particle deposition.
        positions.convert_to_units("code_length")
        morton = compute_morton(
            positions[:, 0],
            positions[:, 1],
            positions[:, 2],
            self.ds.domain_left_edge,
            self.ds.domain_right_edge,
        )
        morton.sort()
        particle_octree = ParticleOctreeContainer(
            [1, 1, 1],
            self.ds.domain_left_edge,
            self.ds.domain_right_edge,
            num_zones=2,
        )
        particle_octree.n_ref = nneighbors * 2
        particle_octree.add(morton)
        particle_octree.finalize()
        pdom_ind = particle_octree.domain_ind(self.selector)
        if fields is None:
            fields = []
        cls = getattr(particle_smooth, f"{method}_smooth", None)
        if cls is None:
            raise YTParticleDepositionNotImplemented(method)
        nz = self.nz
        mdom_ind = self.domain_ind
        nvals = (nz, nz, nz, (mdom_ind >= 0).sum())
        op = cls(nvals, len(fields), nneighbors, kernel_name)
        op.initialize()
        mylog.debug(
            "Smoothing %s particles into %s Octs", positions.shape[0], nvals[-1]
        )
        op.process_particles(
            particle_octree,
            pdom_ind,
            positions,
            fields,
            self.domain_id,
            self._domain_offset,
            self.ds.periodicity,
            self.ds.geometry,
        )
        vals = op.finalize()
        if vals is None:
            return
        if isinstance(vals, list):
            vals = [np.asfortranarray(v) for v in vals]
        else:
            vals = np.asfortranarray(vals)
        return vals

    @cell_count_cache
    def select_icoords(self, dobj):
        return self.oct_handler.icoords(
            dobj.selector, domain_id=self.domain_id, num_cells=self._cell_count
        )

    @cell_count_cache
    def select_fcoords(self, dobj):
        fcoords = self.oct_handler.fcoords(
            dobj.selector, domain_id=self.domain_id, num_cells=self._cell_count
        )
        return self.ds.arr(fcoords, "code_length")

    @cell_count_cache
    def select_fwidth(self, dobj):
        fwidth = self.oct_handler.fwidth(
            dobj.selector, domain_id=self.domain_id, num_cells=self._cell_count
        )
        return self.ds.arr(fwidth, "code_length")

    @cell_count_cache
    def select_ires(self, dobj):
        return self.oct_handler.ires(
            dobj.selector, domain_id=self.domain_id, num_cells=self._cell_count
        )

    def select(self, selector, source, dest, offset):
        n = self.oct_handler.selector_fill(
            selector, source, dest, offset, domain_id=self.domain_id
        )
        return n

    def count(self, selector):
        return -1

    def count_particles(self, selector, x, y, z):
        # We don't cache the selector results
        count = selector.count_points(x, y, z, 0.0)
        return count

    def select_particles(self, selector, x, y, z):
        mask = selector.select_points(x, y, z, 0.0)
        return mask

    def get_vertex_centered_data(self, fields):
        # Make sure the field list has only unique entries
        fields = list(set(fields))
        new_fields = {}
        cg = self.retrieve_ghost_zones(1, fields)
        for field in fields:
            new_fields[field] = cg[field][1:, 1:, 1:].copy()
            np.add(new_fields[field], cg[field][:-1, 1:, 1:], new_fields[field])
            np.add(new_fields[field], cg[field][1:, :-1, 1:], new_fields[field])
            np.add(new_fields[field], cg[field][1:, 1:, :-1], new_fields[field])
            np.add(new_fields[field], cg[field][:-1, 1:, :-1], new_fields[field])
            np.add(new_fields[field], cg[field][1:, :-1, :-1], new_fields[field])
            np.add(new_fields[field], cg[field][:-1, :-1, 1:], new_fields[field])
            np.add(new_fields[field], cg[field][:-1, :-1, :-1], new_fields[field])
            np.multiply(new_fields[field], 0.125, new_fields[field])

        return new_fields


class OctreeSubsetBlockSlicePosition:
    def __init__(self, ind, block_slice):
        self.ind = ind
        self.block_slice = block_slice
        nz = self.block_slice.octree_subset.nz
        self.ActiveDimensions = np.array([nz, nz, nz], dtype="int64")
        self.ds = block_slice.ds

    def __getitem__(self, key):
        bs = self.block_slice
        rv = np.require(
            bs.octree_subset[key][:, :, :, self.ind].T,
            requirements=bs.octree_subset._block_order,
        )
        return rv

    @property
    def id(self):
        return self.ind

    @property
    def Level(self):
        return self.block_slice._ires[0, 0, 0, self.ind]

    @property
    def LeftEdge(self):
        LE = (
            self.block_slice._fcoords[0, 0, 0, self.ind, :].d
            - self.block_slice._fwidth[0, 0, 0, self.ind, :].d * 0.5
        )
        return self.block_slice.octree_subset.ds.arr(
            LE, self.block_slice._fcoords.units
        )

    @property
    def RightEdge(self):
        RE = (
            self.block_slice._fcoords[-1, -1, -1, self.ind, :].d
            + self.block_slice._fwidth[-1, -1, -1, self.ind, :].d * 0.5
        )
        return self.block_slice.octree_subset.ds.arr(
            RE, self.block_slice._fcoords.units
        )

    @property
    def dds(self):
        return self.block_slice._fwidth[0, 0, 0, self.ind, :]

    def clear_data(self):
        pass

    def get_vertex_centered_data(self, fields, smoothed=False, no_ghost=False):
        field = fields[0]
        new_field = self.block_slice.get_vertex_centered_data(fields)[field]
        return {field: new_field[..., self.ind]}

    @contextmanager
    def _field_parameter_state(self, field_parameters):
        yield self.block_slice.octree_subset._field_parameter_state(field_parameters)


class OctreeSubsetBlockSlice:
    def __init__(self, octree_subset, ds):
        self.octree_subset = octree_subset
        self.ds = ds
        self._vertex_centered_data = {}
        # Cache some attributes
        for attr in ["ires", "icoords", "fcoords", "fwidth"]:
            v = getattr(octree_subset, attr)
            setattr(self, f"_{attr}", octree_subset._reshape_vals(v))

    @property
    def octree_subset_with_gz(self):
        subset_with_gz = getattr(self, "_octree_subset_with_gz", None)
        if not subset_with_gz:
            self._octree_subset_with_gz = self.octree_subset.retrieve_ghost_zones(1, [])
        return self._octree_subset_with_gz

    def get_vertex_centered_data(self, fields, smoothed=False, no_ghost=False):
        if no_ghost is True:
            raise NotImplementedError(
                "get_vertex_centered_data without ghost zones for "
                "oct-based datasets has not been implemented."
            )

        # Make sure the field list has only unique entries
        fields = list(set(fields))
        new_fields = {}
        cg = self.octree_subset_with_gz
        for field in fields:
            if field in self._vertex_centered_data:
                new_fields[field] = self._vertex_centered_data[field]
            else:
                finfo = self.ds._get_field_info(field)
                orig_field = cg[field]
                nocts = orig_field.shape[-1]
                new_field = np.zeros((3, 3, 3, nocts), order="F")

                # Compute vertex-centred data as mean of 8 neighbours cell data
                slices = (slice(1, None), slice(None, -1))
                for slx, sly, slz in product(*repeat(slices, 3)):
                    new_field += orig_field[slx, sly, slz]
                new_field *= 0.125

                new_fields[field] = self.ds.arr(new_field, finfo.output_units)

                self._vertex_centered_data[field] = new_fields[field]

        return new_fields

    def __iter__(self):
        for i in range(self._ires.shape[-1]):
            yield i, OctreeSubsetBlockSlicePosition(i, self)


class YTPositionArray(unyt_array):
    @property
    def morton(self):
        self.validate()
        eps = np.finfo(self.dtype).eps
        LE = self.min(axis=0)
        LE -= np.abs(LE) * eps
        RE = self.max(axis=0)
        RE += np.abs(RE) * eps
        morton = compute_morton(self[:, 0], self[:, 1], self[:, 2], LE, RE)
        return morton

    def to_octree(self, num_zones=2, dims=(1, 1, 1), n_ref=64):
        mi = self.morton
        mi.sort()
        eps = np.finfo(self.dtype).eps
        LE = self.min(axis=0)
        LE -= np.abs(LE) * eps
        RE = self.max(axis=0)
        RE += np.abs(RE) * eps
        octree = ParticleOctreeContainer(dims, LE, RE, num_zones=num_zones)
        octree.n_ref = n_ref
        octree.add(mi)
        octree.finalize()
        return octree

    def validate(self):
        if (
            len(self.shape) != 2
            or self.shape[1] != 3
            or self.units.dimensions != length
        ):
            raise YTInvalidPositionArray(self.shape, self.units.dimensions)



import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1 import AxesGrid

import yt

fn = "IsolatedGalaxy/galaxy0030/galaxy0030"
ds = yt.load(fn)  # load data

fig = plt.figure()

# See http://matplotlib.org/mpl_toolkits/axes_grid/api/axes_grid_api.html
# These choices of keyword arguments produce a four panel plot that includes
# four narrow colorbars, one for each plot.  Axes labels are only drawn on the
# bottom left hand plot to avoid repeating information and make the plot less
# cluttered.
grid = AxesGrid(
    fig,
    (0.075, 0.075, 0.85, 0.85),
    nrows_ncols=(2, 2),
    axes_pad=1.0,
    label_mode="1",
    share_all=True,
    cbar_location="right",
    cbar_mode="each",
    cbar_size="3%",
    cbar_pad="0%",
)

fields = [
    ("gas", "density"),
    ("gas", "velocity_x"),
    ("gas", "velocity_y"),
    ("gas", "velocity_magnitude"),
]

# Create the plot.  Since SlicePlot accepts a list of fields, we need only
# do this once.
p = yt.SlicePlot(ds, "z", fields)

# Velocity is going to be both positive and negative, so let's make these
# slices use a linear colorbar scale
p.set_log(("gas", "velocity_x"), False)
p.set_log(("gas", "velocity_y"), False)

p.zoom(2)

# For each plotted field, force the SlicePlot to redraw itself onto the AxesGrid
# axes.
for i, field in enumerate(fields):
    plot = p.plots[field]
    plot.figure = fig
    plot.axes = grid[i].axes
    plot.cax = grid.cbar_axes[i]

# Finally, redraw the plot on the AxesGrid axes.
p.render()

plt.savefig("multiplot_2x2.png")



import numpy as np

from yt.funcs import mylog
from yt.utilities.io_handler import BaseParticleIOHandler


class IOHandlerSDF(BaseParticleIOHandler):
    _dataset_type = "sdf_particles"

    @property
    def _handle(self):
        return self.ds.sdf_container

    def _read_fluid_selection(self, chunks, selector, fields, size):
        raise NotImplementedError

    def _read_particle_coords(self, chunks, ptf):
        chunks = list(chunks)
        data_files = set()
        assert len(ptf) == 1
        assert ptf.keys()[0] == "dark_matter"
        for chunk in chunks:
            for obj in chunk.objs:
                data_files.update(obj.data_files)
        assert len(data_files) == 1
        for _data_file in sorted(data_files, key=lambda x: (x.filename, x.start)):
            yield "dark_matter", (
                self._handle["x"],
                self._handle["y"],
                self._handle["z"],
            ), 0.0

    def _read_particle_fields(self, chunks, ptf, selector):
        chunks = list(chunks)
        data_files = set()
        assert len(ptf) == 1
        assert ptf.keys()[0] == "dark_matter"
        for chunk in chunks:
            for obj in chunk.objs:
                data_files.update(obj.data_files)
        assert len(data_files) == 1
        for _data_file in sorted(data_files, key=lambda x: (x.filename, x.start)):
            for ptype, field_list in sorted(ptf.items()):
                x = self._handle["x"]
                y = self._handle["y"]
                z = self._handle["z"]
                mask = selector.select_points(x, y, z, 0.0)
                del x, y, z
                if mask is None:
                    continue
                for field in field_list:
                    if field == "mass":
                        data = np.ones(mask.sum(), dtype="float64")
                        data *= self.ds.parameters["particle_mass"]
                    else:
                        data = self._handle[field][mask]
                    yield (ptype, field), data

    def _identify_fields(self, data_file):
        fields = [("dark_matter", v) for v in self._handle.keys()]
        fields.append(("dark_matter", "mass"))
        return fields, {}

    def _count_particles(self, data_file):
        pcount = self._handle["x"].size
        if pcount > 1e9:
            mylog.warning(
                "About to load %i particles into memory. "
                "You may want to consider a midx-enabled load",
                pcount,
            )
        return {"dark_matter": pcount}


class IOHandlerHTTPSDF(IOHandlerSDF):
    _dataset_type = "http_sdf_particles"

    def _read_particle_coords(self, chunks, ptf):
        chunks = list(chunks)
        data_files = set()
        assert len(ptf) == 1
        assert ptf.keys()[0] == "dark_matter"
        for chunk in chunks:
            for obj in chunk.objs:
                data_files.update(obj.data_files)
        assert len(data_files) == 1
        for _data_file in data_files:
            pcount = self._handle["x"].size
            yield "dark_matter", (
                self._handle["x"][:pcount],
                self._handle["y"][:pcount],
                self._handle["z"][:pcount],
            ), 0.0

    def _read_particle_fields(self, chunks, ptf, selector):
        chunks = list(chunks)
        data_files = set()
        assert len(ptf) == 1
        assert ptf.keys()[0] == "dark_matter"
        for chunk in chunks:
            for obj in chunk.objs:
                data_files.update(obj.data_files)
        assert len(data_files) == 1
        for _data_file in data_files:
            pcount = self._handle["x"].size
            for ptype, field_list in sorted(ptf.items()):
                x = self._handle["x"][:pcount]
                y = self._handle["y"][:pcount]
                z = self._handle["z"][:pcount]
                mask = selector.select_points(x, y, z, 0.0)
                del x, y, z
                if mask is None:
                    continue
                for field in field_list:
                    if field == "mass":
                        if self.ds.field_info._mass_field is None:
                            pm = 1.0
                            if "particle_mass" in self.ds.parameters:
                                pm = self.ds.parameters["particle_mass"]
                            else:
                                raise RuntimeError
                            data = pm * np.ones(mask.sum(), dtype="float64")
                        else:
                            data = self._handle[self.ds.field_info._mass_field][:][mask]
                    else:
                        data = self._handle[field][:][mask]
                    yield (ptype, field), data

    def _count_particles(self, data_file):
        return {"dark_matter": self._handle["x"].http_array.shape}


class IOHandlerSIndexSDF(IOHandlerSDF):
    _dataset_type = "midx_sdf_particles"

    def _read_particle_coords(self, chunks, ptf):
        dle = self.ds.domain_left_edge.in_units("code_length").d
        dre = self.ds.domain_right_edge.in_units("code_length").d
        for dd in self.ds.midx.iter_bbox_data(dle, dre, ["x", "y", "z"]):
            yield "dark_matter", (dd["x"], dd["y"], dd["z"]), 0.0

    def _read_particle_fields(self, chunks, ptf, selector):
        dle = self.ds.domain_left_edge.in_units("code_length").d
        dre = self.ds.domain_right_edge.in_units("code_length").d
        required_fields = []
        for field_list in sorted(ptf.values()):
            for field in field_list:
                if field == "mass":
                    continue
                required_fields.append(field)

        for dd in self.ds.midx.iter_bbox_data(dle, dre, required_fields):
            for ptype, field_list in sorted(ptf.items()):
                x = dd["x"]
                y = dd["y"]
                z = dd["z"]
                mask = selector.select_points(x, y, z, 0.0)
                del x, y, z
                if mask is None:
                    continue
                for field in field_list:
                    if field == "mass":
                        data = np.ones(mask.sum(), dtype="float64")
                        data *= self.ds.parameters["particle_mass"]
                    else:
                        data = dd[field][mask]
                    yield (ptype, field), data

    def _count_particles(self, data_file):
        dle = self.ds.domain_left_edge.in_units("code_length").d
        dre = self.ds.domain_right_edge.in_units("code_length").d
        pcount_estimate = self.ds.midx.get_nparticles_bbox(dle, dre)
        if pcount_estimate > 1e9:
            mylog.warning(
                "Filtering %i particles to find total. "
                "You may want to reconsider your bounding box.",
                pcount_estimate,
            )
        pcount = 0
        for dd in self.ds.midx.iter_bbox_data(dle, dre, ["x"]):
            pcount += dd["x"].size
        return {"dark_matter": pcount}

    def _identify_fields(self, data_file):
        fields = [("dark_matter", v) for v in self._handle.keys()]
        fields.append(("dark_matter", "mass"))
        return fields, {}


class IOHandlerSIndexHTTPSDF(IOHandlerSIndexSDF):
    _dataset_type = "midx_http_sdf_particles"



import re

import pytest
from unyt import unyt_array
from unyt.exceptions import UnitConversionError

from yt.testing import fake_amr_ds


@pytest.fixture(scope="module")
def reusable_fake_dataset():
    ds = fake_amr_ds(
        fields=[("gas", "density")],
        units=["g/cm**3"],
    )
    return ds


valid_single_str_values = ("center",)
valid_field_loc_str_values = ("min", "max")

DEFAUT_ERROR_MESSAGE = (
    "Expected any of the following\n"
    "- 'c', 'center', 'l', 'left', 'r', 'right', 'm', 'max', or 'min'\n"
    "- a 2 element tuple with 'min' or 'max' as the first element, followed by a field identifier\n"
    "- a 3 element array-like: for a unyt_array, expects length dimensions, otherwise code_lenght is assumed"
)


@pytest.mark.parametrize(
    "user_input",
    (
        # second element can be a single str or a field tuple (2 str), but not three
        (("max", ("not", "a", "field"))),
        # a 1-tuple is also not a valid field key
        (("max", ("notafield",))),
        # both elements need to be str
        (("max", (0, "invalid_field_type"))),
        (("max", ("invalid_field_type", 1))),
    ),
)
def test_invalid_center_type_default_error(reusable_fake_dataset, user_input):
    ds = reusable_fake_dataset
    with pytest.raises(
        TypeError,
        match=re.escape(f"Received {user_input!r}, ")
        + r"but failed to transform to a unyt_array \(obtained .+\)\.",
    ):
        # at the time of writing `axis` is an unused parameter of the base
        # sanitize center method, which is used directly for cartesian coordinate handlers
        # this probably hints that a refactor would make sense to separaet center sanitizing
        # and display_center calculation
        ds.coordinates.sanitize_center(user_input, axis=None)


@pytest.mark.parametrize(
    "user_input, error_type, error_message",
    (
        (
            "bad_str",
            ValueError,
            re.escape(
                "Received unknown center single string value 'bad_str'. "
                + DEFAUT_ERROR_MESSAGE
            ),
        ),
        (
            ("bad_str", ("gas", "density")),
            ValueError,
            re.escape(
                "Received unknown string value 'bad_str'. "
                f"Expected one of {valid_field_loc_str_values} (case insensitive)"
            ),
        ),
        (
            ("bad_str", "density"),
            ValueError,
            re.escape(
                "Received unknown string value 'bad_str'. "
                "Expected one of ('min', 'max') (case insensitive)"
            ),
        ),
        # even with exactly three elements, the dimension should be length
        (
            unyt_array([0.5] * 3, "kg"),
            UnitConversionError,
            "...",  # don't match the exact error message since it's unyt's responsibility
        ),
        # only validate 3 elements unyt_arrays
        (
            unyt_array([0.5] * 2, "cm"),
            TypeError,
            re.escape("Received unyt_array([0.5, 0.5], 'cm')"),
        ),
        (
            unyt_array([0.5] * 4, "cm"),
            TypeError,
            # don't attempt to match error message as details of how
            # a unyt array with more than a couple elements is displayed are out of our control
            "...",
        ),
        (
            # check that the whole shape is used in validation, not just the length (number of rows)
            unyt_array([0.5] * 6, "cm").reshape(3, 2),
            TypeError,
            # don't attempt to match error message as details of how
            # a unyt array with more than a couple elements is displayed are out of our control
            "...",
        ),
    ),
)
def test_invalid_center_special_cases(
    reusable_fake_dataset, user_input, error_type, error_message
):
    ds = reusable_fake_dataset
    with pytest.raises(error_type, match=error_message):
        ds.coordinates.sanitize_center(user_input, axis=None)
